import json
import os
import pickle

import matplotlib.pyplot as plt
import numpy as np

from lofarnn.models.dataloaders.utils import get_lotss_objects


def load_json_arr(json_path):
    lines = []
    with open(json_path, "r") as f:
        for line in f:
            lines.append(json.loads(line))
    return lines


def plot_cutoffs(
    recall_path,
    recall_path_2,
    baseline_path,
    vac_catalog,
    bins=30,
    name="",
    recall_names=["CNN", "Fast RCNN"],
):
    data = pickle.load(open(recall_path, "rb"), fix_imports=True)
    data2 = pickle.load(open(recall_path_2, "rb"), fix_imports=True)
    baseline_data = pickle.load(open(baseline_path, "rb"), fix_imports=True)
    vac_catalog = get_lotss_objects(vac_catalog)
    qual = vac_catalog["LGZ_ID_Qual"]
    pred_source_names = []
    pred_source_recall = []
    pred_source_names2 = []
    pred_source_recall2 = []
    baseline_names = []
    baseline_recalls = []
    # vac_catalog = vac_catalog[vac_catalog["LGZ_Size"] > 15.0]
    # vac_catalog = vac_catalog[vac_catalog["Total_flux"] > 10.0]
    for key in data.keys():
        if key in vac_catalog["Source_Name"].data:
            pred_source_names.append(key)
            pred_source_recall.append(data[key])
    for key in data2.keys():
        if key in vac_catalog["Source_Name"].data:
            pred_source_names2.append(key)
            pred_source_recall2.append(data2[key])
    for key in baseline_data.keys():
        if key in vac_catalog["Source_Name"].data:
            baseline_names.append(key)
            baseline_recalls.append(baseline_data[key])
    pred_source_recall = np.asarray(pred_source_recall)
    pred_source_recall2 = np.asarray(pred_source_recall2)
    baseline_recalls = np.asarray(baseline_recalls)  # [:1630]
    radio_apparent_size = np.zeros(len(pred_source_names))
    radio_apparent_width = np.zeros(len(pred_source_names))
    radio_total_flux = np.zeros(len(pred_source_names))
    radio_z = np.zeros(len(pred_source_names))
    radio_comp = np.zeros(len(pred_source_names))
    quality = np.zeros(len(pred_source_names))
    for i, source_name in enumerate(pred_source_names):
        mask = source_name == vac_catalog["Source_Name"]
        quality[i] = np.nan_to_num(vac_catalog[mask]["LGZ_ID_Qual"].data)
        # get values
        radio_apparent_size[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux[i] = vac_catalog[mask]["Total_flux"].data
        radio_z[i] = vac_catalog[mask]["z_best"].data
        radio_comp[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio = radio_apparent_size / radio_apparent_width
    data_dict = {
        "Axis ratio": radio_axis_ratio,
        "Total flux [mJy]": radio_total_flux,
        "Apparent size [arcsec]": radio_apparent_size,
        "z": radio_z,
        "Number of Components": radio_comp,
        "quality": quality,
    }
    radio_apparent_size3 = np.zeros(len(pred_source_names2))
    radio_apparent_width3 = np.zeros(len(pred_source_names2))
    radio_total_flux3 = np.zeros(len(pred_source_names2))
    radio_z3 = np.zeros(len(pred_source_names2))
    radio_comp3 = np.zeros(len(pred_source_names2))
    for i, source_name in enumerate(pred_source_names2):
        mask = source_name == vac_catalog["Source_Name"]

        # get values
        radio_apparent_size3[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width3[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux3[i] = vac_catalog[mask]["Total_flux"].data
        radio_z3[i] = vac_catalog[mask]["z_best"].data
        radio_comp3[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio3 = radio_apparent_size3 / radio_apparent_width3
    data_dict2 = {
        "Axis ratio": radio_axis_ratio3,
        "Total flux [mJy]": radio_total_flux3,
        "Apparent size [arcsec]": radio_apparent_size3,
        "z": radio_z3,
        "Number of Components": radio_comp3,
    }
    radio_apparent_size2 = np.zeros(len(baseline_names))
    radio_apparent_width2 = np.zeros(len(baseline_names))
    radio_total_flux2 = np.zeros(len(baseline_names))
    radio_z2 = np.zeros(len(baseline_names))
    radio_comp2 = np.zeros(len(baseline_names))
    for i, source_name in enumerate(baseline_names):  # [:1630]):
        mask = source_name == vac_catalog["Source_Name"]

        # get values
        radio_apparent_size2[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width2[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux2[i] = vac_catalog[mask]["Total_flux"].data
        radio_z2[i] = vac_catalog[mask]["z_best"].data
        radio_comp2[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio2 = radio_apparent_size2 / radio_apparent_width2
    baseline_data_dict = {
        "Axis ratio": radio_axis_ratio2,
        "Total flux [mJy]": radio_total_flux2,
        "Apparent size [arcsec]": radio_apparent_size2,
        "z": radio_z2,
        "Number of Components": radio_comp2,
    }
    pred_source_recall = np.array(pred_source_recall)
    pred_source_recall2 = np.array(pred_source_recall2)
    baseline_recalls = np.array(baseline_recalls)

    # Now plot recall cutoffs
    for ylabel in [
        "Apparent size [arcsec]",
        "Total flux [mJy]",
        "Axis ratio",
        "z",
        "Number of Components",
    ]:
        Y = data_dict[ylabel]
        Y2 = baseline_data_dict[ylabel]
        Y3 = data_dict2[ylabel]
        # x_bin_edges = np.linspace(np.nanmin(X)-0.00001, np.nanpercentile(X, 98), bins+1)
        y_bin_edges = np.linspace(
            np.nanpercentile(Y, 1) - 0.00001, np.nanpercentile(Y, 95), bins + 1
        )
        # y_bin_edges = np.linspace(np.nanmin(Y)-0.00001, np.nanpercentile(Y, 98), bins + 1)
        # derive bin centers
        y_bin_width = y_bin_edges[1] - y_bin_edges[0]
        y_bin_centers = y_bin_edges[1:] - y_bin_width / 2
        # now obtain recall
        recall = np.zeros((y_bin_centers.shape[0]))
        recall2 = np.zeros((y_bin_centers.shape[0]))
        recall3 = np.zeros((y_bin_centers.shape[0]))
        qualities = np.zeros((y_bin_centers.shape[0]))
        num_sources = np.zeros((y_bin_centers.shape[0]))
        for j in range(len(y_bin_centers)):
            # get the prediction errors in a bin
            bin_mask = (Y > y_bin_edges[j]) & (Y < y_bin_edges[j + 1])
            bin_contents = pred_source_recall[bin_mask]
            bin_quality = data_dict["quality"][bin_mask]
            bin_contents2 = baseline_recalls[
                (Y2 > y_bin_edges[j]) & (Y2 < y_bin_edges[j + 1])
            ]
            bin_contents3 = pred_source_recall2[
                (Y3 > y_bin_edges[j]) & (Y3 < y_bin_edges[j + 1])
            ]
            # determine recall
            # print(f"Bin Sum: {np.nansum(bin_contents > 0.95)}, len: {len(bin_contents)}")
            recall[j] = np.nansum(bin_contents > 0.95) / len(bin_contents)
            recall2[j] = np.nansum(bin_contents2 > 0.95) / len(bin_contents2)
            recall3[j] = np.nansum(bin_contents3 > 0.95) / len(bin_contents3)
            qualities[j] = np.nansum(bin_quality) / len(bin_quality)
            num_sources[j] = len(bin_contents)
        # Now plot
        fig, (ax3, ax1, ax2) = plt.subplots(
            3, 1, sharex="all", gridspec_kw={"height_ratios": [1, 3, 1], "hspace": 0}
        )
        # gs = fig.add_gridspec(3, hspace=0)
        # ax2, ax1, ax3 = gs.subplots(sharex=True, sharey=False)
        ax1.plot(y_bin_centers, recall, label=f"{recall_names[0]}")
        ax1.plot(y_bin_centers, recall3, label=f"{recall_names[1]}")
        ax1.plot(y_bin_centers, recall2, linestyle="--", label="Baseline")
        fig.suptitle(f"Recall vs {ylabel}")
        ax1.set_ylabel("Recall")
        ax1.set_xlabel(ylabel)
        # ax2 = ax1.twinx()
        ax2.set_xlabel(ylabel)
        ax2.bar(
            y_bin_centers,
            num_sources,
            width=(max(y_bin_centers) - min(y_bin_centers)) / len(y_bin_centers),
            edgecolor="black",
            color="none",
            zorder=10,
        )
        ax2.set_ylabel("# Sources")
        num_sources = num_sources / sum(num_sources)
        ax3.plot(y_bin_centers, np.cumsum(num_sources), c="black")
        ax3.set_ylabel("% Sources")
        ax3.set_xlabel(ylabel)
        if ylabel == "Apparent size [arcsec]":
            ax1.axvline(x=70.0, linestyle="dashed", color="black")
            ax2.axvline(x=70.0, linestyle="dashed", color="black")
            ax3.axvline(x=70.0, linestyle="dashed", color="black")
            mask = y_bin_centers <= 70
            s = (
                f"Baseline: {np.round(np.nanmean(recall2[mask]), 3)}\n"
                f"{recall_names[0]}: {np.round(np.nanmean(recall[mask]), 3)}\n"
                f"{recall_names[1]}: {np.round(np.nanmean(recall3[mask]), 3)}\n "
            )
            ax1.text(x=1, y=0.5, s=s, size=8)
            mask = y_bin_centers > 70
            s = (
                f"Baseline: {np.round(np.nanmean(recall2[mask]), 3)}\n"
                f"{recall_names[0]}: {np.round(np.nanmean(recall[mask]), 3)}\n"
                f"{recall_names[1]}: {np.round(np.nanmean(recall3[mask]), 3)}\n "
            )
            ax1.text(x=72, y=0.5, s=s, size=8)
        ax1.legend(loc="best")
        fig.savefig(f"{ylabel}_Recall_{name}.png", dpi=300)
        plt.clf()
        plt.cla()
        # Now read out recalls for apparent size
        if ylabel == "Apparent size [arcsec]":
            # Size cut
            mask = y_bin_centers <= 70
            print(f"Baseline Mean Recall <= 70 arcseconds: {np.nanmean(recall2[mask])}")
            print(f"CNN Mean Recall <= 70 arcseconds: {np.nanmean(recall[mask])}")
            print(f"FRCNN Mean Recall <= 70 arcseconds: {np.nanmean(recall3[mask])}")
            mask = y_bin_centers > 70
            print(f"Baseline Mean Recall > 70 arcseconds: {np.mean(recall2[mask])}")
            print(f"CNN Mean Recall > 70 arcseconds: {np.mean(recall[mask])}")
            print(f"FRCNN Mean Recall > 70 arcseconds: {np.mean(recall3[mask])}")


def plot_compared_axis_recall(
    recall_path,
    recall_path_2,
    vac_catalog,
    limit,
    jelle_cut=False,
    bins=10,
    output_dir="./",
):
    """
    Plot recall of apparent size to axis ratio
    :param recall_path: Recall path from SourceEvaluator, which has source_name and highest overlap of the limit
    :param vac_catalog: Value-added catalog location
    :param bins: Number of bins for the histogram
    :param limit: str, limit for the recall value for use in saving, title, etc.
    :return:
    """
    data = pickle.load(open(recall_path, "rb"), fix_imports=True)
    data2 = pickle.load(open(recall_path_2, "rb"), fix_imports=True)
    vac_catalog = get_lotss_objects(vac_catalog)
    pred_source_names = []
    pred_source_recall = []
    pred_source_names2 = []
    pred_source_recall2 = []
    if jelle_cut:
        vac_catalog = vac_catalog[vac_catalog["LGZ_Size"] > 15.0]
        vac_catalog = vac_catalog[vac_catalog["Total_flux"] > 10.0]
    for key in data.keys():
        if key in vac_catalog["Source_Name"].data:
            pred_source_names.append(key)
            pred_source_recall.append(data[key])
    for key in data2.keys():
        if key in vac_catalog["Source_Name"].data:
            pred_source_names2.append(key)
            pred_source_recall2.append(data2[key])
    pred_source_recall = np.asarray(pred_source_recall)
    pred_source_recall2 = np.asarray(pred_source_recall2)  # [:1630]
    radio_apparent_size = np.zeros(len(pred_source_names))
    radio_apparent_width = np.zeros(len(pred_source_names))
    radio_total_flux = np.zeros(len(pred_source_names))
    radio_z = np.zeros(len(pred_source_names))
    radio_comp = np.zeros(len(pred_source_names))
    for i, source_name in enumerate(pred_source_names):
        mask = source_name == vac_catalog["Source_Name"]

        # get values
        radio_apparent_size[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux[i] = vac_catalog[mask]["Total_flux"].data
        radio_z[i] = vac_catalog[mask]["z_best"].data
        radio_comp[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio = radio_apparent_size / radio_apparent_width
    data_dict = {
        "Axis ratio": radio_axis_ratio,
        "Total flux [mJy]": radio_total_flux,
        "Apparent size [arcsec]": radio_apparent_size,
        "z": radio_z,
        "Number of Components": radio_comp,
    }
    radio_apparent_size2 = np.zeros(len(pred_source_names2))
    radio_apparent_width2 = np.zeros(len(pred_source_names2))
    radio_total_flux2 = np.zeros(len(pred_source_names2))
    radio_z2 = np.zeros(len(pred_source_names2))
    radio_comp2 = np.zeros(len(pred_source_names2))
    for i, source_name in enumerate(pred_source_names2):  # [:1630]):
        mask = source_name == vac_catalog["Source_Name"]

        # get values
        radio_apparent_size2[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width2[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux2[i] = vac_catalog[mask]["Total_flux"].data
        radio_z2[i] = vac_catalog[mask]["z_best"].data
        radio_comp2[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio2 = radio_apparent_size2 / radio_apparent_width2
    data_dict2 = {
        "Axis ratio": radio_axis_ratio2,
        "Total flux [mJy]": radio_total_flux2,
        "Apparent size [arcsec]": radio_apparent_size2,
        "z": radio_z2,
        "Number of Components": radio_comp2,
    }

    ###calculate recall in bins
    # set which parameters you want to have on the X and Y axis
    for xlabel in ["Apparent size [arcsec]"]:
        for ylabel in ["Total flux [mJy]", "Axis ratio", "z", "Number of Components"]:
            X = data_dict[xlabel]
            Y = data_dict[ylabel]
            X2 = data_dict2[xlabel]  # [:1595]
            Y2 = data_dict2[ylabel]  # [:1595]
            # get edges with maxima determined using percentiles to be robust for outliers
            x_bin_edges = np.linspace(
                np.nanpercentile(X, 1) - 0.00001, np.nanpercentile(X, 98), bins + 1
            )
            # x_bin_edges = np.linspace(np.nanmin(X)-0.00001, np.nanpercentile(X, 98), bins+1)
            y_bin_edges = np.linspace(
                np.nanpercentile(Y, 1) - 0.00001, np.nanpercentile(Y, 95), bins + 1
            )
            # y_bin_edges = np.linspace(np.nanmin(Y)-0.00001, np.nanpercentile(Y, 98), bins + 1)
            # derive bin centers
            x_bin_width = x_bin_edges[1] - x_bin_edges[0]
            x_bin_centers = x_bin_edges[1:] - x_bin_width / 2
            y_bin_width = y_bin_edges[1] - y_bin_edges[0]
            y_bin_centers = y_bin_edges[1:] - y_bin_width / 2

            recall_2D = np.zeros((x_bin_centers.shape[0], y_bin_centers.shape[0]))
            recall_2D2 = np.zeros((x_bin_centers.shape[0], y_bin_centers.shape[0]))
            n_sources = np.zeros(
                (x_bin_centers.shape[0], y_bin_centers.shape[0]), dtype=int
            )
            # now obtain recall
            for i in range(len(x_bin_centers)):
                for j in range(len(y_bin_centers)):
                    # get the prediction errors in a bin
                    bin_contents = pred_source_recall[
                        (X > x_bin_edges[i])
                        & (X < x_bin_edges[i + 1])
                        & (Y > y_bin_edges[j])
                        & (Y < y_bin_edges[j + 1])
                    ]
                    bin_2 = pred_source_recall2[
                        (X2 > x_bin_edges[i])
                        & (X2 < x_bin_edges[i + 1])
                        & (Y2 > y_bin_edges[j])
                        & (Y2 < y_bin_edges[j + 1])
                    ]

                    # determine recall
                    # print(f"Bin Sum: {np.nansum(bin_contents > 0.95)}, len: {len(bin_contents)}")
                    recall_2D[i][j] = np.nansum(bin_contents > 0.95) / len(bin_contents)
                    recall_2D2[i][j] = np.nansum(bin_2 > 0.95) / len(bin_2)
                    # print(f"Recall Model:\n {recall_2D}")
                    # print(f"Recall Baseline:\n {recall_2D2}")
                    # also determine the number of sources
                    n_sources[i][j] = len(bin_contents)
            recall_2D = recall_2D - recall_2D2

            # now get the selection mask
            fig, ax = plt.subplots()

            # get the desired aspect ratio such that the plot is square
            aspectratio = (np.max(x_bin_centers) - np.min(x_bin_centers)) / (
                np.max(y_bin_centers) - np.min(y_bin_centers)
            )
            im = ax.imshow(
                recall_2D.T,
                origin="lower",
                cmap="bwr_r",
                vmin=-1,
                vmax=1,
                zorder=2,
                aspect=aspectratio,
                extent=[
                    np.min(x_bin_edges),
                    np.max(x_bin_edges),
                    np.min(y_bin_edges),
                    np.max(y_bin_edges),
                ],
            )

            xlims = ax.get_xlim()
            ylims = ax.get_ylim()

            # reset view limits
            ax.set_xlim(xlims)
            ax.set_ylim(ylims)

            # indicate the number of sources
            for i in range(n_sources.shape[0]):
                for j in range(n_sources.shape[1]):
                    ax.text(
                        x_bin_centers[i],
                        y_bin_centers[j],
                        str(n_sources[i, j]),
                        ha="center",
                        va="center",
                        color="black",
                        fontsize=5,
                    )

            cbar = plt.colorbar(im, ax=ax)
            cbar.ax.set_ylabel("Recall Improvement Over Baseline")

            ax.set_xticks(x_bin_centers)
            ax.set_yticks(y_bin_centers)

            ax.tick_params(axis="x", labelrotation=40, labelsize="x-small")
            ax.tick_params(axis="y", labelrotation=0, labelsize="x-small")

            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            ax.set_title(f"{limit} recall vs Baseline for {xlabel} vs {ylabel}")
            fig.savefig(
                os.path.join(
                    output_dir, f"{xlabel}-{ylabel}_jelle{jelle_cut}_{limit}.png"
                ),
                dpi=300,
                bbox_inches="tight",
            )
            plt.close()


def plot_axis_recall(
    recall_path, vac_catalog, limit, jelle_cut=False, bins=10, output_dir="./"
):
    """
    Plot recall of apparent size to axis ratio
    :param recall_path: Recall path from SourceEvaluator, which has source_name and highest overlap of the limit
    :param vac_catalog: Value-added catalog location
    :param bins: Number of bins for the histogram
    :param limit: str, limit for the recall value for use in saving, title, etc.
    :return:
    """
    data = pickle.load(open(recall_path, "rb"), fix_imports=True)
    vac_catalog = get_lotss_objects(vac_catalog)
    pred_source_names = []
    pred_source_recall = []
    if jelle_cut:
        vac_catalog = vac_catalog[vac_catalog["LGZ_Size"] > 15.0]
        vac_catalog = vac_catalog[vac_catalog["Total_flux"] > 10.0]
    for key in data.keys():
        if key in vac_catalog["Source_Name"].data:
            pred_source_names.append(key)
            pred_source_recall.append(data[key])
    pred_source_recall = np.asarray(pred_source_recall)
    radio_apparent_size = np.zeros(len(pred_source_names))
    radio_apparent_width = np.zeros(len(pred_source_names))
    radio_total_flux = np.zeros(len(pred_source_names))
    radio_z = np.zeros(len(pred_source_names))
    radio_comp = np.zeros(len(pred_source_names))
    for i, source_name in enumerate(pred_source_names):
        mask = source_name == vac_catalog["Source_Name"]

        # get values
        radio_apparent_size[i] = vac_catalog[mask]["LGZ_Size"].data
        radio_apparent_width[i] = vac_catalog[mask]["LGZ_Width"].data
        radio_total_flux[i] = vac_catalog[mask]["Total_flux"].data
        radio_z[i] = vac_catalog[mask]["z_best"].data
        radio_comp[i] = vac_catalog[mask]["LGZ_Assoc"].data
    radio_axis_ratio = radio_apparent_size / radio_apparent_width
    data_dict = {
        "Axis ratio": radio_axis_ratio,
        "Total flux [mJy]": radio_total_flux,
        "Apparent size [arcsec]": radio_apparent_size,
        "z": radio_z,
        "Number of Components": radio_comp,
    }

    ###calculate recall in bins
    # set which parameters you want to have on the X and Y axis
    for xlabel in ["Apparent size [arcsec]"]:
        for ylabel in ["Total flux [mJy]", "Axis ratio", "z", "Number of Components"]:
            X = data_dict[xlabel]
            Y = data_dict[ylabel]
            # get edges with maxima determined using percentiles to be robust for outliers
            x_bin_edges = np.linspace(
                np.nanpercentile(X, 1) - 0.00001, np.nanpercentile(X, 98), bins + 1
            )
            # x_bin_edges = np.linspace(np.nanmin(X)-0.00001, np.nanpercentile(X, 98), bins+1)
            y_bin_edges = np.linspace(
                np.nanpercentile(Y, 1) - 0.00001, np.nanpercentile(Y, 95), bins + 1
            )
            # y_bin_edges = np.linspace(np.nanmin(Y)-0.00001, np.nanpercentile(Y, 98), bins + 1)
            # derive bin centers
            x_bin_width = x_bin_edges[1] - x_bin_edges[0]
            x_bin_centers = x_bin_edges[1:] - x_bin_width / 2
            y_bin_width = y_bin_edges[1] - y_bin_edges[0]
            y_bin_centers = y_bin_edges[1:] - y_bin_width / 2

            recall_2D = np.zeros((x_bin_centers.shape[0], y_bin_centers.shape[0]))
            n_sources = np.zeros(
                (x_bin_centers.shape[0], y_bin_centers.shape[0]), dtype=int
            )
            # now obtain recall
            for i in range(len(x_bin_centers)):
                for j in range(len(y_bin_centers)):
                    # get the prediction errors in a bin
                    bin_contents = pred_source_recall[
                        (X > x_bin_edges[i])
                        & (X < x_bin_edges[i + 1])
                        & (Y > y_bin_edges[j])
                        & (Y < y_bin_edges[j + 1])
                    ]
                    # determine recall
                    recall_2D[i][j] = np.sum(bin_contents > 0.95) / len(bin_contents)
                    # also determine the number of sources
                    n_sources[i][j] = len(bin_contents)

            # now get the selection mask
            fig, ax = plt.subplots()

            # get the desired aspect ratio such that the plot is square
            aspectratio = (np.max(x_bin_centers) - np.min(x_bin_centers)) / (
                np.max(y_bin_centers) - np.min(y_bin_centers)
            )
            im = ax.imshow(
                recall_2D.T,
                origin="lower",
                cmap="viridis",
                vmin=0,
                vmax=1,
                zorder=2,
                aspect=aspectratio,
                extent=[
                    np.min(x_bin_edges),
                    np.max(x_bin_edges),
                    np.min(y_bin_edges),
                    np.max(y_bin_edges),
                ],
            )

            xlims = ax.get_xlim()
            ylims = ax.get_ylim()

            # reset view limits
            ax.set_xlim(xlims)
            ax.set_ylim(ylims)

            # indicate the number of sources
            for i in range(n_sources.shape[0]):
                for j in range(n_sources.shape[1]):
                    ax.text(
                        x_bin_centers[i],
                        y_bin_centers[j],
                        str(n_sources[i, j]),
                        ha="center",
                        va="center",
                        color="white",
                        fontsize=5,
                    )

            cbar = plt.colorbar(im, ax=ax)
            cbar.ax.set_ylabel("Recall")

            ax.set_xticks(x_bin_centers)
            ax.set_yticks(y_bin_centers)

            ax.tick_params(axis="x", labelrotation=40, labelsize="x-small")
            ax.tick_params(axis="y", labelrotation=0, labelsize="x-small")

            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            ax.set_title(f"Recall for {xlabel} vs {ylabel}, limit: {limit}")
            fig.savefig(
                os.path.join(
                    output_dir, f"{xlabel}-{ylabel}_limit{limit}_jelle{jelle_cut}.png"
                ),
                dpi=200,
                bbox_inches="tight",
            )
            plt.close()


def plot_plots(
    metrics_files,
    experiment_name,
    experiment_dir,
    cuts,
    labels,
    title,
    output_dir,
    colors,
):
    """
    Plot a variety of different metrics, including recall, precision, loss, etc.
    :param metrics_files:
    :param experiment_name:
    :param labels:
    :param output_dir:
    :return:
    """
    metrics_data = []
    for f in metrics_files:
        metrics_data.append(load_json_arr(os.path.join(f)))

    # Now extract the useful information: own recall and precision are the important ones since COCO seems off
    recall = {}
    val_recall = {}
    precision = {}
    val_precision = {}
    iteration = {}
    loss = {}
    val_loss = {}
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"] = []
                precision[
                    f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                ] = []
                val_precision[
                    f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                ] = []
                recall[f"{experiment_name}_train_test/own_recall_{j}/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_{j}/recall"] = []
                precision[f"{experiment_name}_train_test/own_recall_{j}/precision"] = []
                val_precision[f"{experiment_name}_val/own_recall_{j}/precision"] = []

    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                iteration[metrics_files[i]] = []
                loss[metrics_files[i]] = []
                val_loss[metrics_files[i]] = []
                val_recall[f"{experiment_name}_val/own_recall_1/recall"] = []
                val_precision[f"{experiment_name}_val/own_recall_1/precision"] = []
                precision[f"{experiment_name}_train_test/own_recall_1/precision"] = []
                recall[f"{experiment_name}_train_test/own_recall_1/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_1_{cut}/recall"] = []
                val_precision[
                    f"{experiment_name}_val/own_recall_1_{cut}/precision"
                ] = []
                precision[
                    f"{experiment_name}_train_test/own_recall_1_{cut}/precision"
                ] = []
                # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                recall[f"{experiment_name}_train_test/own_recall_1_{cut}/recall"] = []
                for line in metric:
                    try:
                        val_recall[
                            f"{experiment_name}_val/own_recall_{j}_{cut}/recall"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"]
                        )
                        val_precision[
                            f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                            ]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                            ]
                        )
                        recall[
                            f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"
                            ]
                        )
                        val_recall[
                            f"{experiment_name}_val/own_recall_{j}/recall"
                        ].append(line[f"{experiment_name}_val/own_recall_{j}/recall"])
                        val_precision[
                            f"{experiment_name}_val/own_recall_{j}/precision"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{j}/precision"]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_{j}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}/precision"
                            ]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_{j}/recall"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall_{j}/recall"]
                        )
                        val_recall[f"{experiment_name}_val/own_recall_1/recall"].append(
                            line[f"{experiment_name}_val/own_recall/recall"]
                        )
                        val_precision[
                            f"{experiment_name}_val/own_recall_1/precision"
                        ].append(line[f"{experiment_name}_val/own_recall/precision"])
                        precision[
                            f"{experiment_name}_train_test/own_recall_1/precision"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall/precision"]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_1/recall"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall/recall"]
                        )
                        val_recall[
                            f"{experiment_name}_val/own_recall_1_{cut}/recall"
                        ].append(line[f"{experiment_name}_val/own_recall_{cut}/recall"])
                        val_precision[
                            f"{experiment_name}_val/own_recall_1_{cut}/precision"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{cut}/precision"]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_1_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{cut}/precision"
                            ]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_1_{cut}/recall"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{cut}/recall"
                            ]
                        )
                        val_loss[metrics_files[i]].append(line["validation_loss"])
                        loss[metrics_files[i]].append(line["total_loss"])
                        iteration[metrics_files[i]].append(line["iteration"])
                    except:
                        continue

    # Plot the iteration vs loss for the models
    for i, metric in enumerate(metrics_data):
        if "1.0" in labels[i]:
            print(len(iteration[metrics_files[i]]))
        else:
            print(len(iteration[metrics_files[i]]))
        iteration[metrics_files[i]] = iteration[metrics_files[i]][:35]
        loss[metrics_files[i]] = loss[metrics_files[i]][:35]
        val_loss[metrics_files[i]] = val_loss[metrics_files[i]][:35]
        plt.plot(
            iteration[metrics_files[i]],
            loss[metrics_files[i]],
            label=f"{labels[i]} Train",
            color=colors[i],
        )
        plt.plot(
            iteration[metrics_files[i]],
            val_loss[metrics_files[i]],
            linestyle="dashed",
            label=f"{labels[i]} Val",
            color=colors[i],
        )

    plt.legend(loc="lower left")
    plt.title(f"Total Training Loss {title}")
    # plt.xlim(0,200000)
    plt.xlabel("Iteration")
    plt.ylabel("Total Loss")
    plt.yscale("log")
    plt.savefig(os.path.join(output_dir, f"Total_Training_Loss_{title}.png"), dpi=300)
    plt.clf()
    plt.cla()

    # Plot recall for the different cuts for the same models
    for i, metric in enumerate(metrics_data):
        for j in [1, 2, 5, 10, 100]:
            for k, cut in enumerate(cuts):
                plt.plot(
                    iteration[metrics_files[i]],
                    recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"],
                    label=f"{labels[k]} Train",
                    color=colors[k],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"],
                    label=f"{labels[k]} Val",
                    linestyle="dashed",
                    color=colors[k],
                )

            plt.legend(loc="lower right")
            plt.title(f"Recall for limit {j}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Recall")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Recall_limit{j}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot precision for different cuts for same models
    for i, metric in enumerate(metrics_data):
        for j in [1, 2, 5, 10, 100]:
            for k, cut in enumerate(cuts):
                plt.plot(
                    iteration[metrics_files[i]],
                    precision[
                        f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{cut} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_precision[
                        f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{cut} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Precision for limit {j}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Precision")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Precision_limit{j}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot recall and precision based on different limits
    # Plot recall for the different cuts for the same models
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                plt.plot(
                    iteration[metrics_files[i]],
                    recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"],
                    label=f"{j} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"],
                    label=f"{j} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Recall for cut {cut}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Recall")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Recall_cut{cut}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot precision for different cuts for same models
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                plt.plot(
                    iteration[metrics_files[i]],
                    precision[
                        f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{j} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_precision[
                        f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{j} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Precision for cut {cut}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Precision")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Precision_cut{cut}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot recall for different models

    # Plot precision for different models


def plot_combo_plots(
    metrics_files,
    experiment_name,
    experiment_dir,
    cuts,
    labels,
    title,
    output_dir,
    colors,
):
    """
        Plot a variety of different metrics, including recall, precision, loss, etc.
        :param metrics_files:
        :param experiment_name:
        :param labels:
        :param output_dir:
        :return:
        """
    metrics_data = []
    for f in metrics_files:
        metrics_data.append(load_json_arr(os.path.join(experiment_dir, f + ".json")))

    # Now extract the useful information: own recall and precision are the important ones since COCO seems off
    recall = {}
    val_recall = {}
    precision = {}
    val_precision = {}
    iteration = {}
    loss = {}
    val_loss = {}
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"] = []
                precision[
                    f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                ] = []
                val_precision[
                    f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                ] = []
                recall[f"{experiment_name}_train_test/own_recall_{j}/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_{j}/recall"] = []
                precision[f"{experiment_name}_train_test/own_recall_{j}/precision"] = []
                val_precision[f"{experiment_name}_val/own_recall_{j}/precision"] = []

    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [2, 5, 10, 100]:
                iteration[metrics_files[i]] = []
                loss[metrics_files[i]] = []
                val_loss[metrics_files[i]] = []
                val_recall[f"{experiment_name}_val/own_recall_1/recall"] = []
                val_precision[f"{experiment_name}_val/own_recall_1/precision"] = []
                precision[f"{experiment_name}_train_test/own_recall_1/precision"] = []
                recall[f"{experiment_name}_train_test/own_recall_1/recall"] = []
                val_recall[f"{experiment_name}_val/own_recall_1_{cut}/recall"] = []
                val_precision[
                    f"{experiment_name}_val/own_recall_1_{cut}/precision"
                ] = []
                precision[
                    f"{experiment_name}_train_test/own_recall_1_{cut}/precision"
                ] = []
                # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                recall[f"{experiment_name}_train_test/own_recall_1_{cut}/recall"] = []
                for line in metric:
                    try:
                        val_recall[
                            f"{experiment_name}_val/own_recall_{j}_{cut}/recall"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"]
                        )
                        val_precision[
                            f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                            ]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                            ]
                        )
                        recall[
                            f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"
                            ]
                        )
                        val_recall[
                            f"{experiment_name}_val/own_recall_{j}/recall"
                        ].append(line[f"{experiment_name}_val/own_recall_{j}/recall"])
                        val_precision[
                            f"{experiment_name}_val/own_recall_{j}/precision"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{j}/precision"]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_{j}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{j}/precision"
                            ]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_{j}/recall"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall_{j}/recall"]
                        )
                        val_recall[f"{experiment_name}_val/own_recall_1/recall"].append(
                            line[f"{experiment_name}_val/own_recall/recall"]
                        )
                        val_precision[
                            f"{experiment_name}_val/own_recall_1/precision"
                        ].append(line[f"{experiment_name}_val/own_recall/precision"])
                        precision[
                            f"{experiment_name}_train_test/own_recall_1/precision"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall/precision"]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_1/recall"
                        ].append(
                            line[f"{experiment_name}_train_test/own_recall/recall"]
                        )
                        val_recall[
                            f"{experiment_name}_val/own_recall_1_{cut}/recall"
                        ].append(line[f"{experiment_name}_val/own_recall_{cut}/recall"])
                        val_precision[
                            f"{experiment_name}_val/own_recall_1_{cut}/precision"
                        ].append(
                            line[f"{experiment_name}_val/own_recall_{cut}/precision"]
                        )
                        precision[
                            f"{experiment_name}_train_test/own_recall_1_{cut}/precision"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{cut}/precision"
                            ]
                        )
                        # print(line[f"{experiment_name}_train_test/own_recall_{j}/recall"])
                        recall[
                            f"{experiment_name}_train_test/own_recall_1_{cut}/recall"
                        ].append(
                            line[
                                f"{experiment_name}_train_test/own_recall_{cut}/recall"
                            ]
                        )
                        val_loss[metrics_files[i]].append(line["validation_loss"])
                        loss[metrics_files[i]].append(line["total_loss"])
                        iteration[metrics_files[i]].append(line["iteration"])
                    except:
                        continue

    # Plot the iteration vs loss for the models
    for i, metric in enumerate(metrics_data):
        plt.plot(
            iteration[metrics_files[i]],
            loss[metrics_files[i]],
            label=f"{labels[i]} Train",
            color=colors[i],
        )
        plt.plot(
            iteration[metrics_files[i]],
            val_loss[metrics_files[i]],
            linestyle="dashed",
            label=f"{labels[i]} Val",
            color=colors[i],
        )

    plt.legend(loc="upper right")
    plt.title(f"Total Training Loss {title}")
    plt.xlabel("Iteration")
    plt.ylabel("Total Loss")
    plt.yscale("log")
    plt.savefig(os.path.join(output_dir, f"Total_Training_Loss_{title}.png"), dpi=300)
    plt.clf()
    plt.cla()

    # Plot recall for the different cuts for the same models
    for i, metric in enumerate(metrics_data):
        for j in [1, 2, 5, 10, 100]:
            for k, cut in enumerate(cuts):
                plt.plot(
                    iteration[metrics_files[i]],
                    recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"],
                    label=f"{labels[k]} Train",
                    color=colors[k],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"],
                    label=f"{labels[k]} Val",
                    linestyle="dashed",
                    color=colors[k],
                )

            plt.legend(loc="lower right")
            plt.title(f"Recall for limit {j}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Recall")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Recall_limit{j}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot precision for different cuts for same models
    for i, metric in enumerate(metrics_data):
        for j in [1, 2, 5, 10, 100]:
            for k, cut in enumerate(cuts):
                plt.plot(
                    iteration[metrics_files[i]],
                    precision[
                        f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{cut} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_precision[
                        f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{cut} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Precision for limit {j}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Precision")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Precision_limit{j}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot recall and precision based on different limits
    # Plot recall for the different cuts for the same models
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                plt.plot(
                    iteration[metrics_files[i]],
                    recall[f"{experiment_name}_train_test/own_recall_{j}_{cut}/recall"],
                    label=f"{j} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_recall[f"{experiment_name}_val/own_recall_{j}_{cut}/recall"],
                    label=f"{j} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Recall for cut {cut}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Recall")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Recall_cut{cut}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot precision for different cuts for same models
    for i, metric in enumerate(metrics_data):
        for cut in cuts:
            for j in [1, 2, 5, 10, 100]:
                plt.plot(
                    iteration[metrics_files[i]],
                    precision[
                        f"{experiment_name}_train_test/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{j} Train",
                    # color=colors[i],
                )
                plt.plot(
                    iteration[metrics_files[i]],
                    val_precision[
                        f"{experiment_name}_val/own_recall_{j}_{cut}/precision"
                    ],
                    label=f"{j} Val",
                    linestyle="dashed",
                    # color=colors[i],
                )

            plt.legend(loc="lower right")
            plt.title(f"Precision for cut {cut}: {title}, {metrics_files[i]}")
            plt.xlabel("Iteration")
            plt.ylabel("Precision")
            # plt.yscale("log")
            plt.savefig(
                os.path.join(
                    output_dir, f"Precision_cut{cut}_{title}_{metrics_files[i]}.png"
                ),
                dpi=300,
            )
            plt.clf()
            plt.cla()

    # Plot recall for different models

    # Plot precision for different models


def plot_wcs(filename, name, pred, target, aux, comp_catalog):
    """
    Plot the radio images with the WCS for proper RA and DEC in the images
    :param aux: CNN Aux data for this source
    :param name: Name of the source
    """
    from matplotlib import pyplot as plt
    from astropy.coordinates import SkyCoord

    (
        img_array,
        bounding_boxes,
        proposal_boxes,
        sem_seg_five,
        sem_seg_prop_five,
        sem_seg_three,
        sem_seg_prop_three,
        wcs,
    ) = np.load(filename, fix_imports=True, allow_pickle=True)
    ra_array = np.array(aux[1:, 2], dtype=float)
    dec_array = np.array(aux[1:, 3], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")
    pixel_coords = sky_coords.to_pixel(wcs)
    components = comp_catalog[comp_catalog["Source_Name"] == name]
    print(components)
    if len(components) > 1:
        coords = SkyCoord(components["RA"], components["DEC"], unit="deg")
        fluxes = components["Total_flux"].data
        comp_coords = coords.to_pixel(wcs)
        # Now need flux weighted center
        center_x = 0
        center_y = 0
        fluxes = fluxes / np.sum(fluxes)
        for c in range(len(comp_coords[0])):
            center_x += comp_coords[0][c] * fluxes[c]
            center_y += comp_coords[1][c] * fluxes[c]
    else:
        center_x = img_array[:, :, 0].shape[0] / 2
        center_y = img_array[:, :, 0].shape[1] / 2

    # Get Baseline
    source_coord = SkyCoord.from_pixel(xp=center_x, yp=center_y, wcs=wcs)
    d2d = source_coord.separation(sky_coords)
    baseline_pred = np.argmin(d2d.data)  # Need to convert back from Angle to non-Angle
    fig = plt.figure()
    ax = fig.add_subplot(111, projection=wcs)
    im = plt.imshow(
        np.clip(img_array[:, :, 0], 0.0, 1000.0), origin="lower", cmap=plt.cm.viridis
    )
    ax.scatter(pixel_coords, size=2, label="Optical Source")
    ax.scatter(pixel_coords[pred - 1], size=5, c="red", marker="*", label="Predicted")
    ax.scatter(pixel_coords[target - 1], size=5, c="green", marker="x", label="Actual")
    ax.scatter(pixel_coords[1], size=5, c="orange", marker="1", label="Closest")
    ax.scatter(
        pixel_coords[baseline_pred], size=5, c="purple", marker="2", label="Baseline"
    )
    cbar = plt.colorbar(im, ax=ax)
    cbar.ax.set_ylabel("Flux (mJy)")
    plt.title(name)
    plt.xlabel("RA")
    plt.ylabel("Dec")
    plt.legend(loc="best")
    plt.savefig(f"{name}_prediction_plot.png", dpi=300)
    plt.cla()
    plt.clf()



from detectron2.engine.hooks import HookBase
from detectron2.evaluation import inference_context
from detectron2.utils.logger import log_every_n_seconds
from detectron2.data import DatasetMapper, build_detection_test_loader
import detectron2.utils.comm as comm
import numpy as np
import logging
import torch
import time
import datetime


class LossEvalHook(HookBase):
    def __init__(self, eval_period, extra_eval, model, data_loader):
        self._model = model
        self._period = eval_period
        self._data_loader = data_loader
        self._extra_eval = extra_eval

    def _do_loss_eval(self):
        # Copying inference_on_dataset from evaluator.py
        total = len(self._data_loader)
        num_warmup = min(5, total - 1)

        start_time = time.perf_counter()
        total_compute_time = 0
        losses = []
        for idx, inputs in enumerate(self._data_loader):
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_compute_time = 0
            start_compute_time = time.perf_counter()
            if torch.cuda.is_available():
                torch.cuda.synchronize()
            total_compute_time += time.perf_counter() - start_compute_time
            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            seconds_per_img = total_compute_time / iters_after_start
            if idx >= num_warmup * 2 or seconds_per_img > 5:
                total_seconds_per_img = (
                    time.perf_counter() - start_time
                ) / iters_after_start
                eta = datetime.timedelta(
                    seconds=int(total_seconds_per_img * (total - idx - 1))
                )
                log_every_n_seconds(
                    logging.INFO,
                    "Loss on Validation  done {}/{}. {:.4f} s / img. ETA={}".format(
                        idx + 1, total, seconds_per_img, str(eta)
                    ),
                    n=5,
                )
            loss_batch = self._get_loss(inputs)
            losses.append(loss_batch)
        mean_loss = np.mean(losses)
        self.trainer.storage.put_scalar("validation_loss", mean_loss)
        comm.synchronize()

        return losses

    def _get_loss(self, data):
        # How loss is calculated on train_loop
        metrics_dict = self._model(data)
        metrics_dict = {
            k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else float(v)
            for k, v in metrics_dict.items()
        }
        total_losses_reduced = sum(loss for loss in metrics_dict.values())
        return total_losses_reduced

    def after_step(self):
        next_iter = self.trainer.iter + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            self._do_loss_eval()
        elif self._period > 0 and next_iter in self._extra_eval:
            self._do_loss_eval()
        self.trainer.storage.put_scalars(timetest=12)



from setuptools import setup, find_packages

setup(
    name="lofarnn",
    version="0.7.1",
    license="GPLv3",
    author="Jacob Bieker",
    authoer_email="jacob@bieker.tech",
    url="https://github.com/jacobbieker/lofarnn",
    download_url="https://github.com/jacobbieker/lofarnn/archive/v0.7.1.tar.gz",
    keywords=["Radio Astronomy", "PyTorch", "Machine Learning"],
    packages=find_packages(),
    install_requires=["astropy", "numpy", "scikit-image", "pillow", "pytorch", "scipy", "matplotlib"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
)



import os
import pickle
from multiprocessing import Pool, Manager
from pathlib import Path
from typing import Optional, List, Dict, Tuple, Union, Any
from zlib import crc32

import numpy as np
from PIL import Image
from astropy.io import fits
from detectron2.structures import BoxMode

from lofarnn.data.cutouts import convert_to_valid_color, augment_image_and_bboxes
from lofarnn.models.dataloaders.utils import get_lotss_objects
from lofarnn.utils.fits import determine_visible_catalogue_source_and_separation


def mkdirs_safe(directory_list: list):
    """When given a list containing directories,
    checks if these exist, if not creates them."""
    assert isinstance(directory_list, list)
    for directory in directory_list:
        os.makedirs(directory, exist_ok=True)


def create_recursive_directories(prepend_path: str, current_dir: str, dictionary: dict):
    """Give it a base, a current_dir to create and a dictionary of stuff yet to create.
    See create_VOC_style_directory_structure for a use case scenario."""
    mkdirs_safe([os.path.join(prepend_path, current_dir)])
    if dictionary[current_dir] is None:
        return
    else:
        for key in dictionary[current_dir].keys():
            create_recursive_directories(
                os.path.join(prepend_path, current_dir), key, dictionary[current_dir]
            )


def create_coco_style_directory_structure(
    root_directory: str, suffix: str = "", verbose: bool = False
):
    """
     We will create a directory structure identical to that of the CLARAN
     The root directory is the directory in which the directory named 'RGZdevkit' will be placed.
     The structure contained will be as follows:
     LGZ_COCOstyle{suffix}/
        |-- Annotations/
             |-- *.json (Annotation files)
        |-- all/ (train,test,val split directory)
             |-- *.png (Image files)
        |-- train/ (train,test,val split directory)
             |-- *.png (Image files)
        |-- val/ (train,test,val split directory)
             |-- *.png (Image files)
        |-- test/ (train,test,val split directory)
             |-- *.png (Image files)
    """
    directories_to_make = {
        f"COCO{suffix}": {
            "annotations": None,
            "all": None,
            "train": None,
            "val": None,
            "test": None,
        }
    }
    create_recursive_directories(root_directory, f"COCO{suffix}", directories_to_make)
    if verbose:
        print(f"COCO style directory structure created in '{root_directory}'.\n")
    (
        all_directory,
        train_directory,
        val_directory,
        test_directory,
        annotations_directory,
    ) = (
        os.path.join(root_directory, f"COCO{suffix}", "all"),
        os.path.join(root_directory, f"COCO{suffix}", "train"),
        os.path.join(root_directory, f"COCO{suffix}", "val"),
        os.path.join(root_directory, f"COCO{suffix}", "test"),
        os.path.join(root_directory, f"COCO{suffix}", "annotations"),
    )
    return (
        all_directory,
        train_directory,
        val_directory,
        test_directory,
        annotations_directory,
    )


def make_single_cnn_set(
    image_names: List[Path],
    record_list: List[Any],
    set_number: int,
    image_destination_dir: Optional[str],
    pan_wise_location: str = "",
    resize: Optional[Union[int, List[int]]] = None,
    rotation: Optional[Union[List[float], float]] = None,
    convert: bool = True,
    vac_catalog_location: str = "",
    normalize: bool = True,
):
    pan_wise_catalog = fits.open(pan_wise_location, memmap=True)
    pan_wise_catalog = pan_wise_catalog[1].data
    vac_catalog = get_lotss_objects(vac_catalog_location)
    for i, image_name in enumerate(image_names):
        # Get image dimensions and insert them in a python dict
        record_dest_filename = os.path.join(
            image_destination_dir, image_name.stem + f".record.{normalize}.npy"
        )
        if convert:
            image_dest_filename = os.path.join(
                image_destination_dir, image_name.stem + f".cnn.{set_number}.png"
            )
        else:
            if rotation is not None and rotation.any() > 0:
                image_dest_filename = os.path.join(
                    image_destination_dir,
                    image_name.stem + f".cnn.{set_number}.{normalize}.npy",
                )
                record_dest_filename = os.path.join(
                    image_destination_dir,
                    image_name.stem + f".record.{set_number}.{normalize}.npy",
                )
            else:
                image_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".cnn.{normalize}.npy"
                )
        if not os.path.exists(os.path.join(image_dest_filename)):
            (image, cutouts, proposal_boxes, wcs) = np.load(
                image_name, allow_pickle=True
            )  # mmap_mode might allow faster read

            image = np.nan_to_num(image)
            if rotation is not None:
                if isinstance(rotation, (list, tuple, np.ndarray)):
                    (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                        image,
                        cutouts=cutouts,
                        proposal_boxes=proposal_boxes,
                        angle=rotation[set_number],
                        new_size=resize,
                    )
                else:
                    (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                        image,
                        cutouts=cutouts,
                        proposal_boxes=proposal_boxes,
                        angle=np.random.uniform(-rotation, rotation),
                        new_size=resize,
                    )
            else:
                # Need this to convert the bbox coordinates into the correct format
                (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                    image,
                    cutouts=cutouts,
                    proposal_boxes=proposal_boxes,
                    angle=0,
                    new_size=resize,
                    verbose=False,
                )
            width, height, depth = np.shape(image)

            # First R (Radio) channel
            image = image[:, :, 0]
            image_clip = np.copy(image)
            image_none = np.copy(image)
            image = convert_to_valid_color(
                image,
                clip=True,
                lower_clip=0.0,
                upper_clip=1000,
                normalize=normalize,
                scaling="sqrt",
            )
            image_clip = convert_to_valid_color(
                image_clip,
                clip=True,
                lower_clip=0.0,
                upper_clip=1000,
                normalize=normalize,
                scaling=None,
            )
            image_none = convert_to_valid_color(
                image_none,
                clip=False,
                lower_clip=0.0,
                upper_clip=1000,
                normalize=False,
                scaling="sqrt",
            )
            image = np.ma.filled(
                image, fill_value=0.0
            )  # convert back from masked array to normal array
            image_clip = np.ma.filled(
                image_clip, fill_value=0.0
            )  # convert back from masked array to normal array
            image_none = np.ma.filled(
                image_none, fill_value=0.0
            )  # convert back from masked array to normal array
            # Now restack into 3 channel image
            image = np.dstack((image, image_clip, image_none))
            if convert:
                image = np.nan_to_num(image)
                image = (255.0 * image).astype(np.uint8)
                # If converting, only take the first one, radio
                pil_im = Image.fromarray(image, "RGB")
                pil_im.save(image_dest_filename)
            else:
                image = np.nan_to_num(image)  # Only take radio
                np.save(image_dest_filename, image)  # Save to the final destination
        else:
            image = np.load(image_dest_filename)
            height, width, depth = np.shape(image)

        record = {
            "file_name": image_dest_filename,
            "image_id": i,
            "height": height,
            "width": width,
            "depth": depth,
        }
        if not os.path.exists(os.path.join(record_dest_filename)):
            source = vac_catalog[vac_catalog["Source_Name"] == image_name.stem]
            # All optical sources in 150 arcsecond radius of the point
            (
                objects,
                distances,
                angles,
                source_coords,
                sky_coords,
            ) = determine_visible_catalogue_source_and_separation(
                source["RA"], source["DEC"], 150.0 / 3600, pan_wise_catalog
            )
            # Sort from closest to farthest distance
            idx = np.argsort(distances)
            objects = objects[idx]
            distances = distances[idx]
            angles = angles[idx]
            sky_coords = sky_coords[idx]
            layers = [
                "iFApMag",
                "w1Mag",
                "gFApMag",
                "rFApMag",
                "zFApMag",
                "yFApMag",
                "w2Mag",
                "w3Mag",
                "w4Mag",
            ]
            optical_sources = []
            optical_labels = []
            for j, obj in enumerate(objects):
                optical_sources.append([])
                if (
                    obj["objID"] == source["objID"]
                    and obj["AllWISE"] == source["AllWISE"]
                ):
                    optical_labels.append(1)  # Optical Source
                else:
                    optical_labels.append(0)
                optical_sources[-1].append(obj["objID"])
                optical_sources[-1].append(obj["AllWISE"])
                optical_sources[-1].append(obj["ra"])
                optical_sources[-1].append(obj["dec"])
                optical_sources[-1].append(distances[j])
                optical_sources[-1].append(angles[j])
                optical_sources[-1].append(obj["z_best"])
                for layer in layers:
                    value = np.nan_to_num(obj[layer])
                    if normalize:  # Scale to between 0 and 1 for 10 to 28 magnitude
                        value = np.clip(value, 10.0, 28.0)
                        value = (value - 10.0) / (28.0 - 10.0)
                    optical_sources[-1].append(value)
            record["optical_sources"] = optical_sources
            record["optical_labels"] = optical_labels
            record["source_skycoord"] = source_coords
            record["optical_skycoords"] = sky_coords
            if rotation is not None:
                record["rotation"] = rotation[set_number]
            else:
                record["rotation"] = 0.0
            np.save(record_dest_filename, record)
        else:
            record = np.load(record_dest_filename, fix_imports=True, allow_pickle=True)

        # Now add the labels, so need to know which optical source is the true one
        record_list.append(record)


def make_single_coco_annotation_set(
    image_names: List[Path],
    record_list: List[Any],
    set_number: int,
    image_destination_dir: Optional[str],
    multiple_bboxes: bool = True,
    resize: Optional[Union[Tuple[int], int]] = None,
    rotation: Optional[Union[List[float], float]] = None,
    convert: bool = True,
    all_channels: bool = False,
    precomputed_proposals: bool = False,
    normalize: bool = True,
):
    """
    For use with multiprocessing, goes through and does one rotation for the COCO annotations
    :param box_seg: Whether to have segmentation maps and bounding boxes for sources as well as radio components. I.e. bounding boxes with 2 classes:
    Optical source with segmentation of the entire bounding box, and radio component with bounding box of the entire? image and segmentation map inside that
    :param image_names: Image names to load and use for generating dataset
    :param record_list: Array to add the sources to
    :param set_number:
    :param image_destination_dir: The destination directory of the images
    :param multiple_bboxes: Whether to include multiple bounding boxes and segmentation maps
    :param resize: What to resize to
    :param rotation: How much to rotate
    :param convert: Whether to convert to PNG and normalize to between 0 and 255 for all channels
    :param all_channels: Whether to use all 10 channels, or just radio, iband, W1 band
    :param precomputed_proposals: Whether to create precomputed proposals
    :param segmentation: Whether to do segmentation or not, if True, or 5, then uses the 5 sigma segmentation maps, if 3, uses the 3 sigma maps
    :param normalize: Whether to normalize input data between 0 and 1
    :param stats:
    :param cut_size: If to only take center number of pixels, eg. for 200px cutout, 100px in all directrions from center point
    :param verbose:
    :return:
    """
    for i, image_name in enumerate(image_names):
        # Get image dimensions and insert them in a python dict
        if convert:
            image_dest_filename = os.path.join(
                image_destination_dir, image_name.stem + f".{set_number}.png"
            )
        else:
            if rotation is not None and rotation.any() > 0:
                image_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".{set_number}.npy"
                )
            else:
                image_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".npy"
                )
        if multiple_bboxes:
            if rotation is not None and rotation.any() > 0:
                segmap_dest_filename = os.path.join(
                    image_destination_dir,
                    image_name.stem + f".semseg.multi.{set_number}.png",
                )
            else:
                segmap_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".semseg.multi.png"
                )
        else:
            if rotation is not None and rotation.any() > 0:
                segmap_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".semseg.{set_number}.png"
                )
            else:
                segmap_dest_filename = os.path.join(
                    image_destination_dir, image_name.stem + f".semseg.png"
                )
        (image, cutouts, proposal_boxes,) = np.load(
            image_name, allow_pickle=True
        )  # mmap_mode might allow faster read
        image = np.nan_to_num(image)
        # Change order to H,W,C for imgaug
        if rotation is not None:
            if isinstance(rotation, (list, tuple, np.ndarray)):
                (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                    image,
                    cutouts=cutouts,
                    proposal_boxes=proposal_boxes,
                    angle=rotation[set_number],
                    new_size=resize,
                )
            else:
                (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                    image,
                    cutouts=cutouts,
                    proposal_boxes=proposal_boxes,
                    angle=np.random.uniform(-rotation, rotation),
                    new_size=resize,
                )
        else:
            # Need this to convert the bbox coordinates into the correct format
            (image, cutouts, proposal_boxes,) = augment_image_and_bboxes(
                image,
                cutouts=cutouts,
                proposal_boxes=proposal_boxes,
                angle=0,
                new_size=resize,
                verbose=False,
            )
        width, height, depth = np.shape(image)
        # print(segmentation_maps[0].shape)
        if all_channels and depth != 10:
            continue

        # First R (Radio) channel
        image[:, :, 0] = convert_to_valid_color(
            image[:, :, 0],
            clip=True,
            lower_clip=0.0,
            upper_clip=1000,
            normalize=normalize,
            scaling=None,
        )
        for layer in range(1, image.shape[2]):
            image[:, :, layer] = convert_to_valid_color(
                image[:, :, layer],
                clip=True,
                lower_clip=10.0,
                upper_clip=28.0,
                normalize=normalize,
                scaling=None,
            )
        if not os.path.exists(os.path.join(image_dest_filename)):
            if convert:
                image = np.nan_to_num(image)
                image = (255.0 * image).astype(np.uint8)
                # If converting, only take the first three layers, generally Radio, i band, W1 band
                pil_im = Image.fromarray(image[:, :, :3], "RGB")
                pil_im.save(image_dest_filename)
            else:
                image = np.nan_to_num(image)
                np.save(image_dest_filename, image)  # Save to the final destination
        if all_channels:
            rec_depth = 10
        else:
            rec_depth = 3
        record = {
            "file_name": image_dest_filename,
            "image_id": i,
            "height": height,
            "width": width,
            "depth": rec_depth,
        }

        # Insert bounding boxes and their corresponding classes
        objs = []
        # check if there is no optical source
        try:
            if len(cutouts) > 0:  # There is an optical source
                if not multiple_bboxes:
                    cutouts = [
                        cutouts[0]
                    ]  # Only take the first one, the main optical source
                for source_num, bbox in enumerate(cutouts):
                    assert float(bbox[2]) >= float(bbox[0])
                    assert float(bbox[3]) >= float(bbox[1])

                    if bbox[4] == "Other Optical Source":
                        category_id = 0
                    else:
                        category_id = 0
                    obj = {
                        "bbox": [
                            float(bbox[0]),
                            float(bbox[1]),
                            float(bbox[2]),
                            float(bbox[3]),
                        ],
                        "bbox_mode": BoxMode.XYXY_ABS,
                        "category_id": category_id,  # For Optical Source
                        "iscrowd": 0,
                    }
                    objs.append(obj)
        except Exception as e:
            print(e)
            print("No Optical source found")
        if precomputed_proposals:
            record["proposal_boxes"] = proposal_boxes
            record["proposal_objectness_logits"] = np.ones(
                len(proposal_boxes)
            )  # TODO Not sure this is right
            record["proposal_bbox_mode"] = BoxMode.XYXY_ABS
        record["annotations"] = objs
        record_list.append(record)


def create_coco_annotations(
    image_names: List[Path],
    image_destination_dir: Optional[str],
    json_dir: str = "",
    json_name: str = "json_data.pkl",
    multiple_bboxes: bool = True,
    resize: Optional[Union[Tuple[int], int]] = None,
    rotation: Optional[Union[List[float], float]] = None,
    convert: bool = True,
    all_channels: bool = False,
    precomputed_proposals: bool = False,
    segmentation: bool = False,
    normalize: bool = True,
    rotation_names: Optional[List[str]] = None,
    verbose: bool = False,
):
    """
    Creates the annotations for the COCO-style dataset from the npy files available, and saves the images in the correct
    directory
    :param segmentation: Whether to include the segmentation maps or not
    :param image_names: Image names, i.e., the source names
    :param image_destination_dir: The directory the images will end up in
    :param json_dir: The directory where to put the JSON annotation file
    :param json_name: The name of the JSON file
    :param multiple_bboxes: Whether to use multiple bounding boxes, or only the first, for
    example, to only use the main source Optical source, or include others that fall within the
    defined area
    :param rotation: Whether to rotate the images or not, if given as a tuple, it is taken as rotate each image by that amount,
    if a single float, then rotates images randomly between -rotation,rotation 50 times
    :param convert: Whether to convert to PNG files (default), or leave them as NPY files
    :return:
    """

    if rotation is not None:
        if type(rotation) == tuple:
            num_copies = len(rotation)
        else:
            num_copies = 3
    else:
        num_copies = 1
    if rotation_names is not None:
        # Rotate these specific sources ~ 2.5 times more (generally multicomponent ones)
        extra_rotates = []
        t = []
        for i, name in enumerate(image_names):
            print(name.stem)
            print(rotation_names)
            if name.stem in rotation_names:
                extra_rotates.append(i)
                t.append(rotation_names)
        print(f"Matched Names: {t}")
        print(f"Indicies: {extra_rotates} out of {image_names}")
        extra_rotates = np.asarray(extra_rotates)
        image_names = np.asarray(image_names)
        extra_names = image_names[extra_rotates]
        mask = np.ones(len(image_names), np.bool)
        mask[extra_rotates] = 0
        single_names = image_names[mask]
    else:
        single_names = image_names
        extra_names = []
    print(f"Extra Names: {len(extra_names)}")
    print(f"Single Names: {len(single_names)}")
    print(f"Extra Names: {extra_names}")
    print(f"Single Names: {single_names}")
    # List to store single dict for each image
    dataset_dicts = []
    if num_copies > 1:
        manager = Manager()
        pool = Pool(processes=os.cpu_count())
        L = manager.list()
        rotation = np.linspace(0, 180, num_copies)
        [
            pool.apply_async(
                make_single_coco_annotation_set,
                args=[
                    single_names,
                    L,
                    m,
                    image_destination_dir,
                    multiple_bboxes,
                    resize,
                    rotation,
                    convert,
                    all_channels,
                    precomputed_proposals,
                    segmentation,
                    normalize,
                    False,
                    verbose,
                ],
            )
            for m in range(num_copies)
        ]
        print(len(L))
        # Now do the same for the extra copies, but with more rotations, ~2.5 to equal out multi and single comp sources
        num_multi_copies = int(np.ceil(num_copies * 2.5))
        print(f"Num Multi Copies: {num_multi_copies}")
        multi_rotation = np.linspace(0, 180, num_multi_copies)
        [
            pool.apply_async(
                make_single_coco_annotation_set,
                args=[
                    extra_names,
                    L,
                    m,
                    image_destination_dir,
                    multiple_bboxes,
                    resize,
                    multi_rotation,
                    convert,
                    all_channels,
                    precomputed_proposals,
                    segmentation,
                    normalize,
                    False,
                    [],
                    verbose,
                ],
            )
            for m in range(num_multi_copies)
        ]
        pool.close()
        pool.join()
        print(len(L))
        print(f"Length of L: {len(L)}")
        for element in L:
            dataset_dicts.append(element)
        print(f"Length of Dataset Dict: {len(dataset_dicts)}")
        # Write all image dictionaries to file as one json
        json_path = os.path.join(json_dir, json_name)
        with open(json_path, "wb") as outfile:
            pickle.dump(dataset_dicts, outfile)
        if verbose:
            print(f"COCO annotation file created in '{json_dir}'.\n")
        return 0  # Returns to doesnt go through it again

    # Iterate over all cutouts and their objects (which contain bounding boxes and class labels)
    for m in range(num_copies):
        make_single_coco_annotation_set(
            image_names=image_names,
            record_list=dataset_dicts,
            set_number=m,
            image_destination_dir=image_destination_dir,
            multiple_bboxes=multiple_bboxes,
            resize=resize,
            rotation=rotation,
            convert=convert,
            all_channels=all_channels,
            precomputed_proposals=precomputed_proposals,
            normalize=normalize,
        )
    # Write all image dictionaries to file as one json
    json_path = os.path.join(json_dir, json_name)
    with open(json_path, "wb") as outfile:
        pickle.dump(dataset_dicts, outfile)
    if verbose:
        print(f"COCO annotation file created in '{json_dir}'.\n")


def create_cnn_annotations(
    image_names,
    image_destination_dir=None,
    json_dir="",
    json_name="json_data.pkl",
    pan_wise_location="",
    resize=None,
    rotation=None,
    convert=True,
    all_channels=False,
    vac_catalog_location="",
    segmentation=False,
    normalize=True,
    cut_size=None,
    rotation_names=None,
    verbose=False,
):
    """
    Creates the annotations for the COCO-style dataset from the npy files available, and saves the images in the correct
    directory
    :param segmentation: Whether to include the segmentation maps or not
    :param image_names: Image names, i.e., the source names
    :param image_destination_dir: The directory the images will end up in
    :param json_dir: The directory where to put the JSON annotation file
    :param json_name: The name of the JSON file
    :param multiple_bboxes: Whether to use multiple bounding boxes, or only the first, for
    example, to only use the main source Optical source, or include others that fall within the
    defined area
    :param rotation: Whether to rotate the images or not, if given as a tuple, it is taken as rotate each image by that amount,
    if a single float, then rotates images randomly between -rotation,rotation 50 times
    :param convert: Whether to convert to PNG files (default), or leave them as NPY files
    :return:
    """

    if rotation is not None:
        if type(rotation) == tuple:
            num_copies = len(rotation)
        else:
            num_copies = 3
    else:
        num_copies = 1
    if rotation_names is not None:
        # Rotate these specific sources ~ 2.5 times more (generally multicomponent ones)
        extra_rotates = []
        t = []
        for i, name in enumerate(image_names):
            if name.stem in rotation_names:
                extra_rotates.append(i)
                t.append(rotation_names)
        extra_rotates = np.asarray(extra_rotates)
        image_names = np.asarray(image_names)
        extra_names = image_names[extra_rotates]
        mask = np.ones(len(image_names), np.bool)
        mask[extra_rotates] = 0
        single_names = image_names[mask]
    else:
        single_names = image_names
        extra_names = []
    # List to store single dict for each image
    dataset_dicts = []
    if num_copies > 1:
        manager = Manager()
        pool = Pool(processes=os.cpu_count())
        L = manager.list()
        rotation = np.linspace(0, 170, num_copies)
        [
            pool.apply_async(
                make_single_cnn_set,
                args=[
                    single_names,
                    L,
                    m,
                    image_destination_dir,
                    pan_wise_location,
                    resize,
                    rotation,
                    convert,
                    all_channels,
                    vac_catalog_location,
                    segmentation,
                    normalize,
                    False,
                    cut_size,
                    verbose,
                ],
            )
            for m in range(num_copies)
        ]
        print(len(L))
        # Now do the same for the extra copies, but with more rotations, ~2.5 to equal out multi and single comp sources
        num_multi_copies = int(np.ceil(num_copies * 2.5))
        print(f"Num Multi Copies: {num_multi_copies}")
        multi_rotation = np.linspace(0, 170, num_multi_copies)
        [
            pool.apply_async(
                make_single_cnn_set,
                args=[
                    extra_names,
                    L,
                    m,
                    image_destination_dir,
                    pan_wise_location,
                    resize,
                    multi_rotation,
                    convert,
                    all_channels,
                    vac_catalog_location,
                    segmentation,
                    normalize,
                    False,
                    cut_size,
                    verbose,
                ],
            )
            for m in range(num_multi_copies)
        ]
        pool.close()
        pool.join()
        print(len(L))
        print(f"Length of L: {len(L)}")
        for element in L:
            dataset_dicts.append(element)
        print(f"Length of Dataset Dict: {len(dataset_dicts)}")
        # Write all image dictionaries to file as one json
        json_path = os.path.join(json_dir, json_name)
        with open(json_path, "wb") as outfile:
            pickle.dump(dataset_dicts, outfile)
        if verbose:
            print(f"CNN annotation file created in '{json_dir}'.\n")
        return 0  # Returns to doesnt go through it again

    # Iterate over all cutouts and their objects (which contain bounding boxes and class labels)
    bbox_size = []
    # """
    manager = Manager()
    pool = Pool(processes=os.cpu_count())
    L = manager.list()
    rotation = None
    [
        pool.apply_async(
            make_single_cnn_set,
            args=[
                [name],
                L,
                0,
                image_destination_dir,
                pan_wise_location,
                resize,
                None,
                convert,
                all_channels,
                vac_catalog_location,
                segmentation,
                normalize,
                False,
                cut_size,
                verbose,
            ],
        )
        for name in image_names
    ]
    pool.close()
    pool.join()
    print(len(L))
    print(f"Length of L: {len(L)}")
    for element in L:
        dataset_dicts.append(element)
    """
    make_single_cnn_set(
        image_names,
        dataset_dicts,
        0,
        image_destination_dir,
        pan_wise_location,
        resize,
        None,
        convert,
        all_channels,
        vac_catalog_location,
        segmentation,
        normalize,
        False,
        cut_size,
        verbose,
    )
    #"""
    # Write all image dictionaries to file as one json
    json_path = os.path.join(json_dir, json_name)
    with open(json_path, "wb") as outfile:
        pickle.dump(dataset_dicts, outfile)
    if verbose:
        print(f"CNN annotation file created in '{json_dir}'.\n")


def create_cnn_dataset(
    root_directory: str,
    pan_wise_catalog: str = "",
    split_fraction: float = 0.2,
    resize: Optional[Union[Tuple[int], int]] = None,
    rotation: Optional[Union[List[float], float]] = None,
    convert: bool = True,
    all_channels: bool = False,
    vac_catalog: str = "",
    normalize: bool = True,
    subset: str = "",
    multi_rotate_only: Optional[Union[List[str], str]] = None,
    verbose: bool = False,
):
    """
    Create COCO directory structure, if it doesn't already exist, split the image data, and save it to the correct
    directories, and create the COCO annotation file to be loaded into Detectron2, or other similar models
    :param split_fraction: Fraction of the data for the test set. the validation set is rolled into the test set.
    :param root_directory: root directory for the COCO dataset
    :param multiple_bboxes: Whether to include multiple bounding boxes, or only the main source
    :param resize: Image size to resize to, or None if not resizing
    :param convert: Whether to convert npy files to png, or to keep them in the original format, useful for SourceMapper
    :param verbose: Whether to print more data to stdout or not
    :param subset: Whether to limit ones to only the fluxlimit sources, if not empty, should be path to list of source filepaths to use
    :return:
    """

    (
        all_directory,
        train_directory,
        val_directory,
        test_directory,
        annotations_directory,
    ) = create_coco_style_directory_structure(root_directory, verbose=verbose)

    # Gather data from all_directory
    data_split = split_data(
        all_directory, val_split=split_fraction, test_split=split_fraction
    )
    if subset:
        # Keep only those already in the subset
        subset = np.load(subset, allow_pickle=True)
        for d in ["train", "test", "val"]:
            data_split[d] = data_split[d][np.isin(data_split[d], subset)]
        annotations_directory = os.path.join(annotations_directory, "subset")
    if multi_rotate_only:
        l_objects = get_lotss_objects(multi_rotate_only, False)
        # Get all multicomponent sources
        l_objects = l_objects[l_objects["LGZ_Assoc"] > 1]
        multi_names = l_objects["Source_Name"].data
    else:
        multi_names = None
    if len(data_split["val"]) > 0:
        create_cnn_annotations(
            data_split["val"],
            json_dir=annotations_directory,
            image_destination_dir=val_directory,
            json_name=f"cnn_val_norm{normalize}_extra.pkl",
            pan_wise_location=pan_wise_catalog,
            resize=resize,
            rotation=None,
            convert=convert,
            normalize=normalize,
            all_channels=all_channels,
            vac_catalog_location=vac_catalog,
            rotation_names=multi_names,
            verbose=verbose,
        )
    create_cnn_annotations(
        data_split["train"],
        json_dir=annotations_directory,
        image_destination_dir=train_directory,
        json_name=f"cnn_train_test_norm{normalize}_extra.pkl",
        pan_wise_location=pan_wise_catalog,
        resize=resize,
        rotation=None,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        vac_catalog_location=vac_catalog,
        rotation_names=multi_names,
        verbose=verbose,
    )
    create_cnn_annotations(
        data_split["test"],
        json_dir=annotations_directory,
        image_destination_dir=test_directory,
        json_name=f"cnn_test_norm{normalize}_extra.pkl",
        pan_wise_location=pan_wise_catalog,
        resize=resize,
        rotation=None,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        vac_catalog_location=vac_catalog,
        rotation_names=multi_names,
        verbose=verbose,
    )
    create_cnn_annotations(
        data_split["train"],
        json_dir=annotations_directory,
        image_destination_dir=train_directory,
        json_name=f"cnn_train_norm{normalize}_extra.pkl",
        pan_wise_location=pan_wise_catalog,
        resize=resize,
        rotation=rotation,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        vac_catalog_location=vac_catalog,
        rotation_names=multi_names,
        verbose=verbose,
    )


def create_coco_dataset(
    root_directory: str,
    multiple_bboxes: bool = False,
    split_fraction: float = 0.2,
    resize: Optional[Union[Tuple[int], int]] = None,
    rotation: Optional[Union[List[float], float]] = None,
    convert: bool = True,
    all_channels: bool = False,
    precomputed_proposals: bool = False,
    segmentation: bool = False,
    normalize: bool = True,
    subset: str = "",
    multi_rotate_only: Optional[List[str]] = None,
    verbose: bool = False,
):
    """
    Create COCO directory structure, if it doesn't already exist, split the image data, and save it to the correct
    directories, and create the COCO annotation file to be loaded into Detectron2, or other similar models
    :param split_fraction: Fraction of the data for the test set. the validation set is rolled into the test set.
    :param root_directory: root directory for the COCO dataset
    :param multiple_bboxes: Whether to include multiple bounding boxes, or only the main source
    :param resize: Image size to resize to, or None if not resizing
    :param convert: Whether to convert npy files to png, or to keep them in the original format, useful for SourceMapper
    :param verbose: Whether to print more data to stdout or not
    :param subset: Whether to limit ones to only the fluxlimit sources, if not empty, should be path to list of source filepaths to use
    :return:
    """

    (
        all_directory,
        train_directory,
        val_directory,
        test_directory,
        annotations_directory,
    ) = create_coco_style_directory_structure(root_directory, verbose=verbose)

    # Gather data from all_directory
    data_split = split_data(
        all_directory, val_split=split_fraction, test_split=split_fraction
    )
    if subset:
        # Keep only those already in the subset
        subset = np.load(subset, allow_pickle=True)
        for d in ["train", "test", "val"]:
            data_split[d] = data_split[d][np.isin(data_split[d], subset)]
        annotations_directory = os.path.join(annotations_directory, "subset")
    if multi_rotate_only:
        l_objects = get_lotss_objects(multi_rotate_only, False)
        # Get all multicomponent sources
        l_objects = l_objects[l_objects["LGZ_Assoc"] > 1]
        multi_names = l_objects["Source_Name"].data
    else:
        multi_names = None
    create_coco_annotations(
        data_split["train"],
        json_dir=annotations_directory,
        image_destination_dir=train_directory,
        json_name=f"json_train_prop{precomputed_proposals}_all{all_channels}_multi{multiple_bboxes}_seg{segmentation}_norm{normalize}.pkl",
        multiple_bboxes=multiple_bboxes,
        resize=resize,
        rotation=rotation,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        precomputed_proposals=precomputed_proposals,
        rotation_names=multi_names,
        verbose=verbose,
    )
    create_coco_annotations(
        data_split["train"],
        json_dir=annotations_directory,
        image_destination_dir=train_directory,
        json_name=f"json_train_test_prop{precomputed_proposals}_all{all_channels}_multi{multiple_bboxes}_seg{segmentation}_norm{normalize}.pkl",
        multiple_bboxes=multiple_bboxes,
        resize=resize,
        rotation=None,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        precomputed_proposals=precomputed_proposals,
        rotation_names=None,
        verbose=verbose,
    )
    if len(data_split["val"]) > 0:
        create_coco_annotations(
            data_split["val"],
            json_dir=annotations_directory,
            image_destination_dir=val_directory,
            json_name=f"json_val_prop{precomputed_proposals}_all{all_channels}_multi{multiple_bboxes}_seg{segmentation}_norm{normalize}.pkl",
            multiple_bboxes=multiple_bboxes,
            resize=resize,
            rotation=None,
            convert=convert,
            normalize=normalize,
            all_channels=all_channels,
            precomputed_proposals=precomputed_proposals,
            rotation_names=None,
            verbose=verbose,
        )
    create_coco_annotations(
        data_split["test"],
        json_dir=annotations_directory,
        image_destination_dir=test_directory,
        json_name=f"json_test_prop{precomputed_proposals}_all{all_channels}_multi{multiple_bboxes}_seg{segmentation}_norm{normalize}.pkl",
        multiple_bboxes=multiple_bboxes,
        resize=resize,
        rotation=None,
        convert=convert,
        normalize=normalize,
        all_channels=all_channels,
        precomputed_proposals=precomputed_proposals,
        verbose=verbose,
    )


# Function to check test set's identifier.
def test_set_check(identifier: Any, test_ratio: float) -> float:
    return crc32(np.int64(identifier)) & 0xFFFFFFFF < test_ratio * 2 ** 32


# Function to split train/test
def split_train_test_by_id(
    data: Union[np.ndarray, List[str]], test_ratio: float
) -> Tuple[Union[Union[str, List[str]], Any], Union[list, Any]]:
    in_test_set = np.asarray(
        [
            test_set_check(crc32(str(x).split("/")[-1].encode()), test_ratio)
            for x in data
        ]
    )
    return data[~in_test_set], data[in_test_set]


def split_data(
    image_directory: str, val_split: float = 0.2, test_split: float = 0.2
) -> Dict[str, List[str]]:
    """
    Split up the data and return which images should go to which train, test, val directory
    :param image_directory: The directory where all the images are located, i.e. the "all" directory
    :param test_split: Fraction of the data for the test set. the validation set is rolled into the test set.
    :param val_split: Fraction of data in validation set
    :return: A dict containing which images go to which directory
    """

    image_paths = Path(image_directory).rglob("*.npy")
    im_paths = []
    for p in image_paths:
        im_paths.append(p)
    print(len(im_paths))
    train_images, test_images = split_train_test_by_id(
        np.asarray(im_paths), val_split + test_split
    )
    val_images, test_images = split_train_test_by_id(test_images, val_split)
    print(len(train_images))
    print(len(val_images))
    print(len(test_images))
    return {"train": train_images, "val": val_images, "test": test_images}



import multiprocessing
import os
from itertools import repeat
from typing import List, Union, Optional, Tuple

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import skycoord_to_pixel
from scipy.ndimage.filters import gaussian_filter

from lofarnn.models.dataloaders.utils import get_lotss_objects
from lofarnn.utils.coco import create_coco_style_directory_structure
from lofarnn.utils.fits import extract_subimage, determine_visible_catalogue_sources
from lofarnn.visualization.cutouts import plot_three_channel_debug


def pad_with(
    vector: np.ndarray, pad_width: Union[int, List[int]], iaxis: int, **kwargs
):
    """
    Taken from Numpy documentation, will pad with zeros to make lofar image same size as other image
    :param vector:
    :param pad_width:
    :param iaxis:
    :param kwargs:
    :return:
    """
    pad_value = kwargs.get("padder", 0)
    vector[: pad_width[0]] = pad_value
    vector[-pad_width[1] :] = pad_value
    return vector


def make_layer(
    value: float, value_error: float, size: Tuple[int], non_uniform: bool = False
):
    """
    Creates a layer based on the value and the error, if non_uniform is True.

    Designed for adding catalogue data to image stacks

    :param value:
    :param value_error:
    :param size:
    :param non_uniform:
    :return:
    """

    if non_uniform:
        return np.random.normal(value, value_error, size=size)
    else:
        return np.full(shape=size, fill_value=value)


def make_catalogue_layer(
    column_name: str,
    wcs: WCS,
    shape: Union[Tuple[int], int],
    catalogue: Table,
    gaussian: bool = None,
    verbose: bool = False,
):
    """
    Create a layer based off the data in
    :param column_name: Name in catalogue of data to include
    :param shape: Shape of the image data
    :param wcs: WCS of the Radio data, so catalog data can be translated correctly
    :param catalogue: Catalogue to query
    :param gaussian: Whether to smooth the point values with a gaussian
    :return: A Numpy array that holds the information in the correct location
    """

    ra_array = np.array(catalogue["ra"], dtype=float)
    dec_array = np.array(catalogue["dec"], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")

    # Now have the objects, need to convert those RA and Decs to pixel coordinates
    layer = np.zeros(shape=shape)
    coords = skycoord_to_pixel(sky_coords, wcs, 0)
    for index, x in enumerate(coords[0]):
        try:
            if (
                ~np.isnan(catalogue[index][column_name])
                and catalogue[index][column_name] > 0.0
            ):  # Make sure not putting in NaNs
                layer[int(np.floor(coords[0][index]))][
                    int(np.floor(coords[1][index]))
                ] = catalogue[index][column_name]
        except Exception as e:
            if verbose:
                print(f"Failed: {e}")
    if gaussian is not None:
        layer = gaussian_filter(layer, sigma=gaussian)
    return layer


def make_proposal_boxes(wcs: WCS, catalogue: Table, gaussian: Optional[bool] = None):
    """
   Create Faster RCNN proposal boxes for all sources in the image

   The sky_coords seems to be swapped x and y on the boxes, so should be swapped here too
   :param wcs: WCS of the Radio data, so catalog data can be translated correctly
   :param catalogue: Catalogue to query
   :param gaussian: Whether to smooth the point values with a gaussian
   :return: A Numpy array that holds the information in the correct location
   """

    ra_array = np.array(catalogue["ra"], dtype=float)
    dec_array = np.array(catalogue["dec"], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")

    # Now have the objects, need to convert those RA and Decs to pixel coordinates
    proposals = []
    coords = skycoord_to_pixel(sky_coords, wcs, 0)
    for index, x in enumerate(coords[0]):
        try:
            proposals.append(
                make_bounding_box(
                    ra_array[index],
                    dec_array[index],
                    wcs=wcs,
                    class_name="Proposal Box",
                    gaussian=gaussian,
                )
            )
        except Exception as e:
            print(f"Failed Proposal: {e}")
    return proposals


def make_bounding_box(
    ra: Union[float, str],
    dec: Union[float, str],
    wcs: WCS,
    class_name: str = "Optical source",
    gaussian: Optional[bool] = None,
):
    """
    Creates a bounding box and returns it in (xmin, ymin, xmax, ymax, class_name) format
    :param class_name: Class name for the bounding box
    :param ra: RA of the object to make bounding box
    :param dec: Dec of object
    :param wcs: WCS to convert to pixel coordinates
    :param gaussian: Whether gaussian is being used, in which case the box is not int'd but left as a float, and the
    width of the gaussian is used for the width of the bounding box, if it is being used, instead of 'None', it should
    be the width of the Gaussian
    :return: Bounding box coordinates for COCO style annotation
    """
    source_skycoord = SkyCoord(ra, dec, unit="deg")
    box_center = skycoord_to_pixel(source_skycoord, wcs, 0)
    if gaussian is None:
        # Now create box, which will be accomplished by taking int to get xmin, ymin, and int + 1 for xmax, ymax
        xmin = int(np.floor(box_center[0])) - 0.5
        ymin = int(np.floor(box_center[1])) - 0.5
        ymax = ymin + 1
        xmax = xmin + 1
    else:
        xmin = int(np.floor(box_center[0])) - gaussian
        ymin = int(np.floor(box_center[1])) - gaussian
        xmax = np.ceil(int(np.floor(box_center[0])) + gaussian)
        ymax = np.ceil(int(np.floor(box_center[1])) + gaussian)

    return xmin, ymin, xmax, ymax, class_name, box_center


def create_cutouts(
    mosaic: Union[str, List[str], set],
    value_added_catalog: Union[Table, str],
    pan_wise_catalog: Union[Table, str],
    component_catalog: Union[Table, str],
    mosaic_location: str,
    save_cutout_directory: str,
    gaussian: Optional[bool] = None,
    all_channels: Optional[bool] = False,
    source_size: Optional[bool] = None,
    verbose: Optional[bool] = False,
):
    """
    Create cutouts of all sources in a field

    :param mosaic: Name of the field to use
    :param value_added_catalog: The VAC of the LoTSS data release
    :param pan_wise_catalog: The PanSTARRS-ALLWISE catalogue used for Williams, 2018, the LoTSS III paper
    :param mosaic_location: The location of the LoTSS DR2 mosaics
    :param save_cutout_directory: Where to save the cutout npy files
    :param all_channels: Whether to include all possible channels (grizy,W1,2,3,4 bands) in npy file or just (radio,i,W1)
    :param fixed_size: Whether to use fixed size cutouts, in arcseconds, or the LGZ size (default: LGZ)
    :param verbose: Whether to print extra information or not
    :return:
    """
    lofar_data_location = os.path.join(mosaic_location, mosaic, "mosaic-blanked.fits")
    lofar_rms_location = os.path.join(mosaic_location, mosaic, "mosaic.rms.fits")
    if gaussian is False:
        gaussian = None
    if type(pan_wise_catalog) == str:
        print("Trying To Open")
        pan_wise_catalog = fits.open(pan_wise_catalog, memmap=True)
        pan_wise_catalog = pan_wise_catalog[1].data
        print("Opened Catalog")
    # Load the data once, then do multiple cutouts
    try:
        fits.open(lofar_data_location, memmap=True)
        fits.open(lofar_rms_location, memmap=True)
    except:
        if verbose:
            print(f"Mosaic {mosaic} does not exist!")

    mosaic_cutouts = value_added_catalog[value_added_catalog["Mosaic_ID"] == mosaic]
    # Go through each cutout for that mosaic
    for l, source in enumerate(mosaic_cutouts):
        if not os.path.exists(
            os.path.join(save_cutout_directory, f"{source['Source_Name']}.npy")
        ):
            img_array = []
            # Get the ra and dec of the radio source
            source_ra = source["RA"]
            source_dec = source["DEC"]
            # Get the size of the cutout needed
            if source_size is None or source_size is False:
                source_size = (
                    source["LGZ_Size"] * 1.5
                ) / 3600.0  # in arcseconds converted to archours
                source_size *= np.sqrt(2)
            print(source_size)
            try:
                lhdu = extract_subimage(
                    lofar_data_location,
                    source_ra,
                    source_dec,
                    source_size,
                    verbose=verbose,
                )
            except:
                if verbose:
                    print(
                        f"Failed to make data cutout for source: {source['Source_Name']}"
                    )
                continue
            try:
                lrms = extract_subimage(
                    lofar_rms_location,
                    source_ra,
                    source_dec,
                    source_size,
                    verbose=verbose,
                )
            except:
                if verbose:
                    print(
                        f"Failed to make rms cutout for source: {source['Source_Name']}"
                    )
                continue
            img_array.append(lhdu[0].data / lrms[0].data)  # Makes the Radio/RMS channel
            header = lhdu[0].header
            wcs = WCS(header)
            # Now change source_size to size of cutout, or root(2)*source_size so all possible sources are included
            # exit()

            # Now time to get the data from the catalogue and add that in their own channels
            if verbose:
                print(f"Image Shape: {img_array[0].data.shape}")
            # Should now be in Radio/RMS, i, W1 format, else we skip it
            # Need from catalog ra, dec, iFApMag, w1Mag, also have a z_best, which might or might not be available for all
            if all_channels:
                layers = [
                    "iFApMag",
                    "w1Mag",
                    "gFApMag",
                    "rFApMag",
                    "zFApMag",
                    "yFApMag",
                    "w2Mag",
                    "w3Mag",
                    "w4Mag",
                ]
            else:
                layers = ["iFApMag", "w1Mag"]
            # Get the catalog sources once, to speed things up
            # cuts size in two to only get sources that fall within the cutout, instead of ones that go twice as large
            cutout_catalog = determine_visible_catalogue_sources(
                source_ra, source_dec, source_size / 2, pan_wise_catalog
            )
            # Now determine if there are other sources in the area
            other_visible_sources = determine_visible_catalogue_sources(
                source_ra, source_dec, source_size / 2, mosaic_cutouts
            )

            # Now make proposal boxes
            proposal_boxes = np.asarray(
                make_proposal_boxes(
                    wcs, img_array[0].shape, cutout_catalog, gaussian=gaussian
                )
            )
            for layer in layers:
                tmp = make_catalogue_layer(
                    layer, wcs, img_array[0].shape, cutout_catalog, gaussian=gaussian
                )
                img_array.append(tmp)

            img_array = np.array(img_array)
            if verbose:
                print(img_array.shape)
            img_array = np.moveaxis(img_array, 0, 2)
            # Include another array giving the bounding box for the source
            bounding_boxes = []
            try:
                source_bbox = make_bounding_box(
                    source["ID_ra"], source["ID_dec"], wcs, gaussian=gaussian
                )
                assert source_bbox[1] >= 0
                assert source_bbox[0] >= 0
                assert source_bbox[3] < img_array.shape[0]
                assert source_bbox[2] < img_array.shape[1]
                source_bounding_box = list(source_bbox)
                bounding_boxes.append(source_bounding_box)
            except:
                print("Source not in bounds")
                continue
            if verbose:
                plot_three_channel_debug(
                    img_array, bounding_boxes, 1, bounding_boxes[0][5]
                )
            # Now segmentation map
            source_components = component_catalog[
                component_catalog["Source_Name"] == source["Source_Name"]
            ]
            # Now go through and for any other sources in the field of view, add those
            for other_source in other_visible_sources:
                other_components = component_catalog[
                    component_catalog["Source_Name"] == other_source["Source_Name"]
                ]
                other_bbox = make_bounding_box(
                    other_source["ID_ra"],
                    other_source["ID_dec"],
                    wcs,
                    class_name="Other Optical Source",
                    gaussian=gaussian,
                )
                if ~np.isclose(other_bbox[0], bounding_boxes[0][0]) and ~np.isclose(
                    other_bbox[1], bounding_boxes[0][1]
                ):  # Make sure not same one
                    if (
                        other_bbox[1] >= 0
                        and other_bbox[0] >= 0
                        and other_bbox[3] < img_array.shape[0]
                        and other_bbox[2] < img_array.shape[1]
                    ):
                        bounding_boxes.append(
                            list(other_bbox)
                        )  # Only add the bounding box if it is within the image shape

            # Now save out the combined file
            bounding_boxes = np.array(bounding_boxes)

            # Save out the IDs of the cutout catalog and other catalog, for reverse indexing into optical catalogs from
            # results

            if verbose:
                print(bounding_boxes)
            combined_array = [
                img_array,
                bounding_boxes,
                proposal_boxes,
                wcs,
            ]
            try:
                np.save(
                    os.path.join(save_cutout_directory, source["Source_Name"]),
                    combined_array,
                )
            except Exception as e:
                if verbose:
                    print(f"Failed to save: {e}")
        else:
            print(f"Skipped: {l}")


def create_source_dataset(
    cutout_directory: str,
    pan_wise_location: str,
    value_added_catalog_location: str,
    component_catalog_location: str,
    dr_two_location: str,
    gaussian: Optional[bool] = None,
    all_channels: Optional[bool] = False,
    fixed_size: Optional[Union[int, float]] = None,
    filter_lgz: bool = True,
    verbose: bool = False,
    use_multiprocessing: bool = False,
    strict_filter: bool = False,
    filter_optical: bool = True,
    no_source: bool = False,
    num_threads: Optional[int] = os.cpu_count(),
):
    """

    :param cutout_directory: Directory to store the cutouts
    :param pan_wise_location: The location of the PanSTARRS-ALLWISE catalog
    :param value_added_catalog_location: Location of the LoTSS Value Added Catalog
    :param dr_two_location: The location of the LoTSS DR2 Mosaic Locations
    :param gaussian: Whether to spread out data layers with Gaussian of specified width
    :param use_multiprocessing: Whether to use multiprocessing
    :param num_threads: Number of threads to use, if multiprocessing is true
    :param strict_filter: Use the same filtering as for Jelle's subsample, with total flux > 10 mJy, and size > 15 arcseconds
    :param filter_optical: Whether to filter out sources with only optical sources or not
    :param filter_lgz: Whether to filter on LGZ_Size
    :return:
    """
    l_objects = get_lotss_objects(value_added_catalog_location, False)
    print(len(l_objects))
    if filter_lgz:
        l_objects = l_objects[~np.isnan(l_objects["LGZ_Size"])]
        print(len(l_objects))
    if filter_optical:
        if no_source:
            l_objects = l_objects[np.isnan(l_objects["ID_ra"])]
            l_objects = l_objects[np.isnan(l_objects["ID_dec"])]
        else:
            l_objects = l_objects[~np.isnan(l_objects["ID_ra"])]
            l_objects = l_objects[~np.isnan(l_objects["ID_dec"])]
        print(len(l_objects))
    else:  # Otherwise, remove those with optical IDs
        l_objects = l_objects[np.isnan(l_objects["ID_ra"])]
        l_objects = l_objects[np.isnan(l_objects["ID_dec"])]
        print(len(l_objects))
    if strict_filter:
        l_objects = l_objects[l_objects["LGZ_Size"] > 15.0]
        l_objects = l_objects[l_objects["Total_flux"] > 10.0]
        print(len(l_objects))
    mosaic_names = set(l_objects["Mosaic_ID"])

    comp_catalog = get_lotss_objects(component_catalog_location, False)

    # Go through each object, creating the cutout and saving to a directory
    # Create a directory structure identical for detectron2
    (
        all_directory,
        train_directory,
        val_directory,
        test_directory,
        annotations_directory,
    ) = create_coco_style_directory_structure(cutout_directory)

    # Now go through each source in l_objects and create a cutout of the fits file
    # Open the Panstarrs and WISE catalogue
    if fixed_size is False:
        fixed_size = None

    if use_multiprocessing:
        pool = multiprocessing.Pool(num_threads)
        pool.starmap(
            create_cutouts,
            zip(
                mosaic_names,
                repeat(l_objects),
                repeat(pan_wise_location),
                repeat(comp_catalog),
                repeat(dr_two_location),
                repeat(all_directory),
                repeat(gaussian),
                repeat(all_channels),
                repeat(fixed_size),
                repeat(verbose),
            ),
        )
    else:
        create_cutouts(
            mosaic=mosaic_names,
            value_added_catalog=l_objects,
            pan_wise_catalog=pan_wise_location,
            component_catalog=comp_catalog,
            mosaic_location=dr_two_location,
            save_cutout_directory=all_directory,
            gaussian=gaussian,
            all_channels=all_channels,
            source_size=fixed_size,
            verbose=verbose,
        )



import pickle
from typing import List, Union

import numpy as np
import torch
from torch.utils.data import Dataset


class RadioSourceDataset(Dataset):
    """Radio Source dataset."""

    def __init__(
        self,
        json_file: Union[str, List[str]],
        single_source_per_img: bool = True,
        num_sources: int = 40,
        shuffle: bool = False,
        norm: bool = True,
        transform=None,
        remove_no_source: bool = True,
        fraction: float = 1.0,
    ):
        """
        Args:
            json_file (string): Path to the json file with annotations
            single_source_per_img (bool, optional): Whether to give all sources with an image, or a single source per image
        """
        if isinstance(json_file, str):
            self.annotations = pickle.load(open(json_file, "rb"), fix_imports=True)
        else:
            self.annotations = pickle.load(open(json_file[0], "rb"), fix_imports=True)
            for f in json_file[1:]:
                self.annotations.extend(pickle.load(open(f, "rb"), fix_imports=True))
        if fraction < 1.0:
            import random

            random.shuffle(self.annotations)
            self.annotations = self.annotations[: int(len(self.annotations) * fraction)]
        self.norm = norm
        self.transform = transform
        # Remove any non-standard files
        print(f"Len Anno: {len(self.annotations)}")
        print(f"JSON File: {json_file}")
        new_anno = []
        for anno in self.annotations:
            if isinstance(anno, np.ndarray):
                anno = anno.item()
            if anno["height"] == anno["width"] == 200:
                if (
                    remove_no_source
                    and np.count_nonzero(anno["optical_labels"][:num_sources]) != 0
                ):
                    # Check if there is a source within the cutoff, if not, ignore it as well, unless we want non source ones
                    new_anno.append(anno)
                else:
                    new_anno.append(anno)
        self.annotations = new_anno
        print(f"Len Anno After Purge: {len(self.annotations)}")

        self.single_source = single_source_per_img
        self.num_sources = num_sources
        self.shuffle = shuffle

        # for length and indexing in if single optical source per image
        self.mapping = {}
        if self.single_source:
            total = 0
            for i, annotation in enumerate(self.annotations):
                for j, _ in enumerate(annotation["optical_sources"]):
                    if j < self.num_sources:
                        self.mapping[total] = (i, j)
                        total += 1
            self.length = total
        else:
            self.length = len(self.annotations)
        print(f"Total Items: {self.length}")

    def __len__(self):
        return self.length

    def load_single_source(self, idx):
        """
        Given a single index, get the single source, image, and label for it
        """
        anno = self.annotations[self.mapping[idx][0]]
        image = np.load(anno["file_name"], fix_imports=True)
        image = image.reshape((1, image.shape[0], image.shape[1]))
        image = torch.from_numpy(image).float()
        source = anno["optical_sources"][self.mapping[idx][1]]
        source[0] = source[0].value / (0.03)  # Distance (arcseconds)
        source[1] = source[1].value / (2 * np.pi)  # Convert to between 0 and 1
        source[2] = source[2] / 7.0  # Redshift
        source = np.asarray(source)
        label = anno["optical_labels"][self.mapping[idx][1]]
        # First one is Optical, second one is Not
        if label:
            label = np.array([1, 0])  # True
        else:
            label = np.array([0, 1])  # False
        if self.transform:
            image, source = self.transform(image, source)
        return {
            "images": image,
            "sources": torch.from_numpy(source).float(),
            "labels": torch.from_numpy(label).float(),
            "names": self._get_source_name(anno["file_name"]),
        }

    @staticmethod
    def _get_source_name(name):
        source_name = name.split("/")[-1].split(".cnn")[0]
        return source_name

    def load_multi_source(self, idx):
        """
        Given single index, get all the sources and labels, shuffling the order
        """
        anno = self.annotations[idx]
        image = np.load(
            anno["file_name"].replace("/run/media/jacob/T7", "/home/jacob"),
            fix_imports=True,
        )
        image = image.reshape((1, anno["height"], anno["width"]))
        # print(image.shape)
        for i, item in enumerate(anno["optical_sources"]):
            anno["optical_sources"][i][0] = (
                anno["optical_sources"][i][0].value - 0.0
            ) / (
                0.03 - 0.0
            )  # Distance (arcseconds) for first 40 elements
            anno["optical_sources"][i][1] = anno["optical_sources"][i][1].value / (
                2 * np.pi
            )
            anno["optical_sources"][i][2] = (
                np.clip(anno["optical_sources"][i][2], 0.0, 7.0) - 0.0
            ) / (
                7.0 - 0.0
            )  # Redshift
            if self.norm:
                for j in range(3, len(anno["optical_sources"][i])):
                    value = anno["optical_sources"][i][j]
                    value = np.clip(value, 10.0, 28.0)
                    anno["optical_sources"][i][j] = (value - 10.0) / (28.0 - 10.0)
        anno["optical_sources"].insert(
            0, [0 for _ in range(len(anno["optical_sources"][0]))]
        )
        if np.max(anno["optical_sources"][: (self.num_sources - 1)]) == 0:
            anno["optical_labels"].insert(
                0, 1
            )  # Give a 1 if no source in cutoff location out one for No Source
        else:
            anno["optical_labels"].insert(0, 0)
        sources = np.asarray(anno["optical_sources"])
        labels = np.asarray(anno["optical_labels"])

        # Shuffling order of sources and labels, and only taking x number of sources
        sources = sources[: self.num_sources]
        labels = labels[: self.num_sources]
        if self.shuffle:
            indices = np.indices(labels.shape)
            np.random.shuffle(indices)
            sources = sources[indices]
            labels = labels[indices]
            labels = labels.reshape(self.num_sources)
        else:
            sources = sources.reshape(1, sources.shape[0], sources.shape[1])
        image = torch.from_numpy(image).float()
        if self.transform:
            image, sources = self.transform(image, sources)
        return {
            "images": image,
            "sources": torch.from_numpy(sources).float(),
            "labels": torch.from_numpy(labels).float(),
            "names": self._get_source_name(anno["file_name"]),
        }

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        if self.single_source:
            return self.load_single_source(idx)
        else:
            return self.load_multi_source(idx)


def collate_variable_fn(batch):
    images = []
    names = []
    max_size = 0
    for item in batch:
        names.append(item["names"])
        if (
            item["images"].shape[-1] > max_size
        ):  # last element should be either width or height, so good for this
            max_size = item["images"].shape[-1]

    if max_size > 400:
        max_size = 400

    # Second time to pad out tensors for this
    for item in batch:
        if item["images"].shape[-1] != max_size:
            images.append(
                torch.squeeze(
                    torch.nn.functional.interpolate(
                        item["images"].unsqueeze_(0), (max_size, max_size)
                    ),
                    0,
                )
            )
        else:
            images.append(item["images"])

    images = torch.stack(images, dim=0)
    data = torch.stack([item["sources"] for item in batch], dim=0)
    target = torch.stack([item["labels"] for item in batch], dim=0)
    return {"images": images, "sources": data, "labels": target, "names": names}


def collate_autokeras_fn(batch):
    images = []
    names = []
    max_size = 0
    for item in batch:
        names.append(item["names"])
        if (
            item["images"].shape[-1] > max_size
        ):  # last element should be either width or height, so good for this
            max_size = item["images"].shape[-1]

    if max_size > 400:
        max_size = 400

    # Second time to pad out tensors for this
    for item in batch:
        if item["images"].shape[-1] != max_size:
            images.append(
                torch.squeeze(
                    torch.nn.functional.interpolate(
                        item["images"].unsqueeze_(0), (max_size, max_size)
                    ),
                )
            )
        else:
            images.append(item["images"])

    images = torch.stack(images, dim=0)
    data = torch.stack([item["sources"] for item in batch], dim=0).squeeze_()
    target = torch.stack([item["labels"] for item in batch], dim=0)
    return {"images": images, "sources": data, "labels": target, "names": names}



import numpy as np
import torch
import torch.nn as nn

"""

Set of baseline models to compare the other models against, includes:

1. Model which only chooses the closest point
2. Model that chooses the flux-weighted closest point, using the inverse of the total flux
3. Model that chooses the flux-weighted closest point, based on the combined flux of W1 and i-band

All models are designed to work with the multisource setup, to speed up the process

"""


def closest_point_model(data):
    return torch.from_numpy(np.array([[1]]))


class ClosestPointModel(nn.Module):
    def __init__(self, num_image_layers, num_sources, config):
        super(ClosestPointModel, self).__init__()

    def forward(self, image, data):
        # Return the label of 1, as its the closest point in the non-shuffled multi
        result = np.array([[1]])
        return torch.from_numpy(result)


class FluxWeightedModel(nn.Module):
    def __init__(self):
        super(FluxWeightedModel, self).__init__()

    def forward(self, image, data):

        # Sum up the last 9 elements as the total flux
        fluxes = data[:, 3:]
        distances = data[:, 0]
        total_fluxes = np.sum(fluxes, axis=1)
        values = distances / total_fluxes
        # Remove first empty one
        values = values[1:]
        # Return the smallest value, so smallest distance * 1/total flux
        return torch.from_numpy(np.argmin(values))
